"""
Helper constructs for parsing the ARM instruction set.
This module will be imported along with 'from mwcp.utils import construct'
and accessible from the submodule "ARM". (e.g. construct.ARM.LDR)
"""

from . import version28 as construct
from .version28 import this

from . import helpers
from mwcp.utils import elffileutils


def _ByteSwapped(subcon, **ctx):
    r"""
    MODIFIED version of ByteSwapped that allows providing a context.
    Swap the byte order within boundaries of the given subcon.

    :param subcon: the subcon on top of byte swapped bytes
    :param **ctx: Context passed to subcon.sizeof()

    Example::

        Int24ul <--> ByteSwapped(Int24ub)
    """
    size = subcon.sizeof(**ctx)
    return construct.Transformed(subcon, construct.swapbytes, size, construct.swapbytes, size)


# Single Data Transfer (LDR, STR)
_ldr_str_inst = construct.BitStruct(
    'cond' / construct.Nibble,
    construct.Const(1, construct.BitsInteger(2)),  # must be '01'
    'reg_imm_offset' / construct.Bit,              # 0 = immediate offset, 1 = register offset
    'pre_post_indexing' / construct.Bit,           # 0 = post, 1 = pre
    'up_down' / construct.Bit,                     # 0 = down, 1 = up
    'byte_word' / construct.Bit,                   # 0 = word, 1 = byte
    'write_back' / construct.Flag,
    'load_store' / construct.Bit,                  # 0 = store, 1 = load
    'base_register' / construct.Nibble,
    'src_dest_register' / construct.Nibble,
    'offset' / construct.IfThenElse(
        this.reg_imm_offset,
        construct.Octet >> construct.Nibble,       # shift applied to Rm >> Rm
        construct.BitsInteger(12)
    )
)

LDR = construct.ExprValidator(_ByteSwapped(_ldr_str_inst, reg_imm_offset=0), this.load_store == 1)


# Data Processing
_data_proc_inst = construct.BitStruct(
    'cond' / construct.Nibble,
    construct.Const(0, construct.BitsInteger(2)),  # must be '00'
    'reg_imm_operand' / construct.Bit,             # 0 = immediate, 1 = register
    'opcode' / construct.Enum(
        construct.Nibble,
        AND=0x0, EOR=0x1, SUB=0x2, RSB=0x3, ADD=0x4, ADC=0x5, SBC=0x6, RSC=0x7,
        TST=0x8, TEQ=0x9, CMP=0xA, CMN=0xB, ORR=0xC, MOV=0xD, BIC=0xE, MVN=0xF,
    ),
    'set_cond' / construct.Flag,
    'operand_1_reg' / construct.Nibble,
    'dest_reg' / construct.Nibble,
    'operand_2' / construct.IfThenElse(
        this.reg_imm_operand,
        construct.Octet >> construct.Nibble,       # shift applied to Rm >> Rm
        construct.Nibble >> construct.Octet,       # rotate applied to Imm >> Imm
    ),
)
# TODO: Finish adding support for analyzing data processing instructions.
# (shifting/rotating will need to applied to the second operand)


def ELFPointer(inst, inst_end, subcon, elf=None):
    r"""
    This is the ARM version of ELFPointer.
    This subconstruct takes two arguments which
    specify the parsed ARM instruction containing an immediate offset in its second operand
    and the end offset (physical) for said instruction.

    The following ARM instructions are currently supported:
        - LDR

    Example: for the instruction "LDR  R1, =data_offset"
    spec = Struct(
        'inst' / ARM.LDR,
        'inst_end' / Tell,
        'data' / ARM.ELFPointer(this.inst, this.inst_end, Bytes(100))
    )

    spec = Struct(
        're' / Regex(
            '\x01\x03(?P<data_ldr_inst>.{4})(?P<end>)\x06\x07', data_ldr_inst=ARM.LDR, end=Tell),
        'data' / ARM.ELFPointer(this.re.data_ldr_inst, this.re.end, Bytes(100))
    )

    spec.parse(file_data, elf=elf_object)

    :param inst: a construct.Container or function that represents the assembly instruction
    :param inst_end: an int or a function that represents the location of the end of the instruction.
    :param subcon: the subcon to use at the offset
    :param elf: Optional elftools.ELFFile file object.
        (if not supplied here, this must be supplied during parse()/build()
    """
    def _obtain_literal_pool_mem_offset(ctx):
        """Obtains the memory offset to the entry in the literal pool."""
        # Validate LDR instruction
        _inst = inst(ctx._) if callable(inst) else inst
        if _inst.load_store != 1:
            raise construct.ConstructError('Load/Store bit must be set to 1')
        if _inst.base_register != 15 or _inst.reg_imm_offset == 1:
            raise construct.ConstructError(
                'Only instructions with PC relative addressing is currently supported.')
        if _inst.write_back:
            raise construct.ConstructError('Write back cannot be enabled for PC relative addressing.')
        # According to spec, PC is an address 8 bytes from the start of the instruction.
        # (Which means 4 bytes from end.)
        _elf = elf or ctx._params.elf
        _inst_end = inst_end(ctx._) if callable(inst_end) else inst_end
        _inst_end = elffileutils.obtain_memory_offset(_inst_end, elf=_elf)
        pc = _inst_end + 4
        mem_offset = pc + _inst.offset
        return mem_offset

    # HACK: FocusLast (which is FocusedSeq) will try to create a child context when it performs it's parsing.
    # The user will be unaware of this shift and can cause issues if the subcon is dynamic.
    # Therefore, patch the given subcon to use the parent context during parsing.
    # TODO: Embedded() should allow for this functionality!
    class _Embedded(construct.Subconstruct):
        def _parse(self, stream, context, path):
            return self.subcon._parsereport(stream, context._, path)
    subcon = _Embedded(subcon)

    # Use original ELFPointer to create a pointer to the entry in the literal pool, which
    # in turn, is a pointer to the data we actually want.
    return helpers.FocusLast(
        helpers.ELFPointer(_obtain_literal_pool_mem_offset, construct.Int32ul, elf=elf),
        helpers.ELFPointer(this[0], subcon, elf=elf),
    )



"""This is an example parser used to show the different methods of adding data to the reporter."""
import logging
import os

from mwcp import Parser

logger = logging.getLogger(__name__)


class Foo(Parser):
    DESCRIPTION = "Foo"

    @classmethod
    def identify(cls, file_object):
        # identifies if the parser can parse the given file.
        return True

    def run(self):
        # retrieve input file
        input_file = self.file_object

        # standardized metadata
        self.reporter.add_metadata("url", u"http://127.0.0.1")

        # demonstrate access to sample
        logger.info("size of inputfile is {} bytes".format(len(input_file.file_data)))

        # other, non-standardized metadata
        # also demonstrate use of pefile object
        if input_file.pe:
            self.reporter.add_metadata("other", {"section0": input_file.pe.sections[0].Name.rstrip(b"\x00")})

        # demonstrate file output
        self.reporter.output_file(b"hello world", "fooconfigtest.txt", "example output file")

        # demonstrate use of filename()
        logger.info("operating on inputfile {}".format(input_file.file_name))

        # demonstrate use of managed tempdir
        with open(os.path.join(self.reporter.managed_tempdir, "footmp.txt"), "w") as f:
            f.write(
                "This is a temp file created in a directory that will be managed by the mwcp framework. \
                The directory will initially be empty, so there is no worry about name collisions. \
                The directory is deleted after this module run ends, unless tempcleanup is disabled."
            )



r"""
poshdeob - Powershell Deobfuscator

This utility is used for converting obfuscated code and pulling strings along the way.

WARNING: This is a very rudimentary parser that doesn't support a lot of things.
 It makes no promises in deobfuscating all your code.

Usage:
    >>> from mwcp.utils import poshdeob

    # Deobfuscate strings
    >>> deobfuscated, strings = poshdeob.deobfuscate('''\
    ... ${iP} = ("{0}{2}{6}{3}{5}{4}{1}" -f'htt','8000','p://192.1','.','0:','10','68.200')
    ... ${m} = ("{1}{0}{2}"-f'VsbG8','aG','=')
    ... ${r} = (('@*'+'{1'+'}(){0}['+'_')-f [ChaR]124,[ChaR]36)
    ... ${P} = ((("{1}{4}{3}{2}{0}"-f 'e{0}','HK','twar','Sof','CU:{0}'))  -f [cHAR]92)
    ... ${S} = ((("{0}{3}{2}{4}{1}"-f'HKL','Y5a',':Y5aSoftwa','M','re'))  -CrepLAcE ([CHar]89+[CHar]53+[CHar]97),[CHar]92)
    ... ${data} = '%1101000T1100101<1101100<1101100&1101111'-spLIT 'U'-sPlit 'T'-SPliT'%'-spLiT'j' -sPLIT'&' -SpLiT '<'
    ... ${data2} = 'G1101000N1100101m1101100%1101100E1101111'.split( '>G{m:%~NqE')
    ... ${secret} = $ShELlID[1]+$shelLiD[13]+'x'
    ... ''')

    >>> print(deobfuscated)
    ${iP} = 'http://192.168.200.100:8000'
    ${m} = 'aGVsbG8='
    ${r} = '@*$()|[_'
    ${P} = 'HKCU:\Software\'
    ${S} = 'HKLM:\Software\'
    ${data} = ('', '1101000', '1100101', '1101100', '1101100', '1101111')
    ${data2} = ('', '1101000', '1100101', '1101100', '1101100', '1101111')
    ${secret} = 'iex'
    <BLANKLINE>

    >>> strings
    ['http://192.168.200.100:8000', 'aGVsbG8=', '@*$()|[_', 'HKCU:\\Software\\', 'HKLM:\\Software\\', '', '1101000', '1100101', '1101100', '1101100', '1101111', '', '1101000', '1100101', '1101100', '1101100', '1101111', 'iex']

    # Embedded strings can also be processed recursively
    >>> deobfuscated, strings = poshdeob.deobfuscate('''\
    ... ${i`P} = ("{0}{2}{6}{3}{5}{4}{1}" -f'htt','8000','p://192.1','.','0:','10','68.200')
    ... ${data} = "
    ...     ${secret} = -join $ShELlID[1,13]+'x'
    ...     ${r} = (('@*'+'{1'+'}(){0}['+'_')-f [ChaR]124,[ChaR]36)
    ... "
    ... ''', recursive=True)
    >>> print(deobfuscated)
    ${i`P} = 'http://192.168.200.100:8000'
    ${data} = "
        ${secret} = 'iex'
        ${r} = '@*$()|[_'
    "
    <BLANKLINE>
    >>> strings
    ['http://192.168.200.100:8000', 'iex', '@*$()|[_']
"""

import argparse
from typing import Tuple, List

import pyparsing as pp
import re
import sys


_PARSER = None


# region PARSING HOOKS


# Some variables found to be used in obfuscated code with the most probable values.
# TODO: Fill this up!
_VARIABLE_LOOKUP = {
    "pshome": r"C:\Windows\System32\WindowsPowerShell\v1.0",
    "shellid": "Microsoft.PowerShell",
    "env:public": r"C:\Users\Public",
    "env:comspec": r"C:\Windows\system32\cmd.exe",
    "verbosepreference.tostring()": "SilentlyContinue",
}


def _indexing(tokens):
    indices = tokens.indices
    if len(indices) == 1:
        return tokens.data[int(indices[0])]
    else:
        return [tokens.data[int(i)] for i in indices]


def _string_format(tokens):
    format_string = tokens.format_string
    for format in tokens.format:
        format_string = format_string.format(*format.params)
    return format_string


def _string_replace(tokens):
    data = tokens.data
    for replace in tokens.replace:
        # Only escape "\" otherwise regex complains if it trailing.
        old = replace.old.replace("\\", "\\\\")
        new = replace.new.replace("\\", "\\\\")
        data = re.sub(old, new, data, flags=re.IGNORECASE if replace.command != "creplace" else 0)
    return data


def _split(tokens):
    split_data = [tokens.data]
    for split in tokens.split:
        for delimiter in split.delimiters:
            new_split_data = []
            for data in split_data:
                new_split_data.extend(re.split(delimiter, data))
            split_data = new_split_data
    return split_data


# endregion


# region PARSING GRAMMER


def OptionalParen(expr, parenthesis="()"):
    """
    Wraps pyparsing expression to add optional parenthesis.

    >>> hello = OptionalParen(pp.Literal('hello'))
    >>> hello.parseString('hello')
    (['hello'], {})
    >>> hello.parseString('((hello))')
    (['hello'], {})
    >>> goodbye = OptionalParen(pp.Literal('goodbye'), parenthesis='{}')
    >>> goodbye.parseString('goodbye')
    (['goodbye'], {})
    >>> goodbye.parseString('{goodbye}')
    (['goodbye'], {})
    """
    lpar, rpar = map(pp.Suppress, parenthesis)
    term = pp.Forward()
    term <<= expr | (lpar + term + rpar)
    return term


def _gen_parser():
    r"""
    Generates PyParsing grammar for parsing common powershell operations.

    Tests:
    >>> parser = _gen_parser()
    >>> parser.parseString("'{1} {0}'-f 'world','hello'")
    (['hello world'], {})
    >>> parser.parseString('''
    ...     'fGshellolNRfGs'-rEplaCE  ((([cHaR]108+[cHaR]78+[cHaR]82))),'!' .rePLace('fGs',[cHaR]96)''')
    (['`hello!`'], {})
    >>> parser.parseString("'ATBZCFD'-spLIT 'Z'-SPLIT'T'  -spLiT 'F'")
    (['A', 'B', 'C', 'D'], {})
    >>> parser.parseString("$ENv:PuBlIc[13]")
    (['i'], {})
    >>> parser.parseString("('h', 'e', 'l', 'lo')-JOIn ''")
    (['hello'], {})
    >>> parser.parseString("'he`llo'")
    (['hello'], {})
    >>> parser.parseString("'FOtestingFO'.RePLaCE('FO','`')")
    (['`testing`'], {})
    """
    char = ("[" + pp.CaselessKeyword("char") + "]" + pp.Word(pp.nums)("num")).setParseAction(lambda t: chr(int(t.num)))
    string = (
        (pp.Suppress("'") + "`" + pp.Suppress("'"))
        | (pp.Suppress('"') + "`" + pp.Suppress('"'))
        | pp.QuotedString("'", escChar="`", escQuote="''", multiline=True, convertWhitespaceEscapes=False)
        | pp.QuotedString('"', escChar="`", escQuote='""', multiline=True, convertWhitespaceEscapes=False)
    )
    variable = ("$" + pp.oneOf(_VARIABLE_LOOKUP.keys(), caseless=True)("var")).setParseAction(
        lambda t: _VARIABLE_LOOKUP[t.var.lower()]
    )

    _string = pp.Suppress(pp.Optional("[" + pp.CaselessKeyword("string") + "]")) + OptionalParen(
        pp.Suppress(pp.Optional("[" + pp.CaselessKeyword("string") + "]")) + string | char | variable
    )
    concat_string = OptionalParen(
        pp.delimitedList(OptionalParen(_string), delim="+").setParseAction(lambda t: "".join(t))
    )

    # TODO: Support ranges and other fancy indexing.
    indexing = (concat_string("data") + "[" + pp.delimitedList(pp.Word(pp.nums))("indices") + "]").setParseAction(
        _indexing
    )

    # Combine used to enforce there is no space between "c" and "replace"
    _replace_command = pp.Combine(pp.Optional(pp.CaselessLiteral("c")) + pp.CaselessKeyword("replace"))("command")
    string_replace = (
        concat_string("data")
        + pp.OneOrMore(
            pp.Group(
                (pp.Combine("-" + _replace_command) + concat_string("old") + "," + concat_string("new"))
                | (
                    "."
                    + ("'" + _replace_command + "'" | '"' + _replace_command + '"' | _replace_command)
                    + ("(" + concat_string("old") + "," + concat_string("new") + ")")
                )
            )
        )("replace")
    ).setParseAction(_string_replace)

    string_format = (
        concat_string("format_string")
        + pp.OneOrMore(pp.Group(pp.CaselessKeyword("-f") + pp.delimitedList(concat_string)("params")))("format")
    ).setParseAction(_string_format)

    split = (
        concat_string("data")
        + pp.OneOrMore(
            pp.Group(
                (pp.CaselessKeyword("-split") + concat_string("delimiters"))
                | ("." + pp.CaselessKeyword("split") + "(" + concat_string("delimiters") + ")")
            )
        )("split")
    ).setParseAction(_split)

    join = (
        OptionalParen(pp.delimitedList(concat_string)("string_list"))
        + pp.CaselessKeyword("-join")
        + concat_string("join_string")
    ).setParseAction(lambda t: t.join_string.join(t.string_list))

    join_unary = (
        (pp.CaselessKeyword("-join") | pp.CaselessKeyword("[string]::join"))
        + "("
        + OptionalParen(pp.delimitedList(concat_string)("string_list"))
        + ")"
    ).setParseAction(lambda t: "".join(t.string_list))

    # fmt: off
    poss_elements = OptionalParen(
        string_format
        | string_replace
        | split
        | join_unary
        | join
        | indexing
        | concat_string
    )
    # fmt: on

    return poss_elements


# endregion


def _format_code_string(string):
    """Formats string into user readable string that can be placed into powershell code."""
    # Use least used quotes for best readability.
    if string.count("'") > string.count('"'):
        code_string = '"' + string.replace('"', '""') + '"'
    else:
        code_string = "'" + string.replace("'", "''") + "'"
    return code_string


def deobfuscate(code, depth=32, recursive=False) -> Tuple[str, List[str]]:
    """
    Deobfuscates strings found in powershell code.

    :param code: obfuscated powershell code
    :param depth: Number of levels of deobfuscation to run. (defaults to 32)
    :param recursive: Whether to recursively deobfuscate found strings.
                      (be careful, this can be very slow!!)

    returns: tuple containing -- (deobfuscate code, list of strings found)
    """
    if depth <= 0:
        raise ValueError("Depth must be a positive number.")
    orig_depth = depth

    if isinstance(code, bytes):
        code = code.decode("latin1")

    # Generate parser on first run.
    global _PARSER
    if not _PARSER:
        _PARSER = _gen_parser()
        _PARSER.keepTabs = True  # start/end offsets get screwy if we don't enable this.

    # Continuously run code through string deobfuscation until we don't get anything new.
    # (This is necessary because pyparsing is not a true recursive descent parser)
    prev_code = ""
    strings = []
    while depth and prev_code != code:
        depth -= 1
        prev_code = code
        strings = []
        code_replacements = []
        for result, start, end in _PARSER.scanString(code):
            if recursive:
                new_result = []
                for string in result:
                    deob_code, sub_strings = deobfuscate(string, depth=orig_depth, recursive=True)
                    new_result.append(deob_code)
                    strings.extend(sub_strings or [deob_code])
                result = new_result
            else:
                strings.extend(result)

            # replace obfuscated code with less obfuscated code.
            code_string = ", ".join(map(_format_code_string, result))
            # Only wrap parenthesis if more than one string.
            if len(result) > 1:
                code_string = "({})".format(code_string)

            code_replacements.append((start, end, code_string))

        # Replace code with new code.
        new_code = ""
        index = 0
        for start, end, code_string in code_replacements:
            # Sometimes pyparsing includes whitespace at the end of the parsed string for some reason...
            try:
                while code[end - 1] in ("\r", "\n", " "):
                    end -= 1
            except IndexError:
                pass
            new_code += code[index:start] + code_string
            index = end
        new_code += code[index:]
        code = new_code

    return code, strings


def main():
    """CLI interface"""
    arg_parser = argparse.ArgumentParser("Powershell Deobfuscator")
    arg_parser.add_argument("INPUT", help="Input file (or code) to deobfuscate")
    arg_parser.add_argument("OUTPUT", nargs="?", help="Deobfuscated file (default: stdout)")

    args = arg_parser.parse_args()

    if args.OUTPUT:
        output = open(args.OUTPUT, "w")
    else:
        output = sys.stdout

    try:
        with open(args.INPUT, "r") as fo:
            deob_code, _ = deobfuscate(fo.read())
            output.write(deob_code)
    finally:
        if args.OUTPUT:
            output.close()


if __name__ == "__main__":
    main()



"""
Helper methods for setting up multiprocessing workers with logging capabilities
"""

import logging
import multiprocessing as mp
import multiprocessing.pool

logger = logging.getLogger(__name__)

from mwcp import registry
from mwcp.utils import logutil


def initializer(parser_sources, default_source):
    """Initializer function that runs at the beginning of each process creation."""
    registry._sources = parser_sources  # Propagate registered parser information.
    registry._default_source = default_source


class TProcess(mp.Process):
    """
    Slighted modified subclass of :class:`multiprocessing.Process`.

    Use this in place of ``Process`` to enable logging in the spawned process.
    """

    def __init__(self, group=None, target=None, name=None, args=(), kwargs=None):
        kwargs = kwargs or {}
        # NOTE: Forcing group to be None since BaseProcess asserts it to be None.
        super(TProcess, self).__init__(group=None, target=target, name=name, args=args, kwargs=kwargs)
        self.queue = logutil.mp_queue

    def run(self):
        logutil.setup_logging(queue=self.queue)
        logger.debug("Setup logger in {}".format(mp.current_process().name))
        super(TProcess, self).run()


class TPool(mp.pool.Pool):
    """
    Version of :class:`multiprocessing.pool.Pool` that uses :class:`TProcess`.
    """

    Process = TProcess

    def __init__(self, processes=None, maxtasksperchild=None):
        """Overwrite to add initializer."""
        super(TPool, self).__init__(
            processes=processes,
            maxtasksperchild=maxtasksperchild,
            initializer=initializer,
            initargs=(registry._sources, registry._default_source),
        )



"""Stores default configuration values."""

import logging
import os
import pkg_resources


import appdirs
from ruamel.yaml import YAML


logger = logging.getLogger(__name__)
yaml = YAML()


class Config(dict):

    CONFIG_FILE_NAME = 'config.yml'
    USER_CONFIG_DIR = appdirs.user_config_dir('mwcp', appauthor=False)

    # Fields which contain a file or directory path.
    PATH_FIELDS = ['LOG_CONFIG_PATH', 'TESTCASE_DIR', 'MALWARE_REPO', 'PARSER_DIR', 'PARSER_CONFIG_PATH']

    def __repr__(self):
        return 'Config({})'.format(super(Config, self).__repr__())

    @property
    def user_path(self):
        """Returns the path to the user config file."""
        # Get user directory.
        cfg_dir = self.USER_CONFIG_DIR
        if not os.path.isdir(cfg_dir):
            os.makedirs(cfg_dir)

        # Create a user copy if it doesn't exist.
        cfg_file_path = os.path.join(cfg_dir, self.CONFIG_FILE_NAME)
        if not os.path.isfile(cfg_file_path):
            with pkg_resources.resource_stream('mwcp.config', self.CONFIG_FILE_NAME) as default_cfg:
                with open(cfg_file_path, 'wb') as fp:
                    fp.write(default_cfg.read())

        # Also copy over log_config.yml
        log_config_path = os.path.join(cfg_dir, 'log_config.yml')
        if not os.path.isfile(log_config_path):
            default_log_cfg = pkg_resources.resource_stream('mwcp.config', 'log_config.yml')
            with open(log_config_path, 'wb') as fp:
                fp.write(default_log_cfg.read())
            default_log_cfg.close()

        return cfg_file_path

    def load(self, file_path=None):
        """Loads configuration file."""
        if not file_path:
            file_path = self.user_path

        with open(file_path, 'r') as fp:
            config = dict(yaml.load(fp))

        # Convert file path into absolute paths.
        directory = os.path.dirname(file_path)
        for key, value in config.items():
            if key in self.PATH_FIELDS:
                value = os.path.expanduser(value)
                value = os.path.expandvars(value)
                value = os.path.join(directory, value)
                value = os.path.abspath(value)
                config[key] = value
        self.update(config)


_config = Config()

# We are going to manually add the fields.json path because
# the fields.json file is not currently designed to be modified.
_config['FIELDS_PATH'] = os.path.abspath(pkg_resources.resource_filename('mwcp.config', 'fields.json'))




"""Utilities for setting up logging."""
import copy
import errno
import logging.config
import logging.handlers
import multiprocessing as mp
import os
import platform
import sys
import threading
import traceback
import warnings
from collections import deque

import appdirs
import yaml

import mwcp

# Queue used to send over log messages from child to main process.
# (See mwcp.utils.multi_proc for its use.)
mp_queue = mp.Queue()


class LevelCharFilter(logging.Filter):
    """Logging filter used to add a 'level_char' format variable."""

    def filter(self, record):
        if record.levelno >= logging.ERROR:
            record.level_char = "!"
        elif record.levelno >= logging.WARN:
            record.level_char = "-"
        elif record.levelno >= logging.INFO:
            record.level_char = "+"
        elif record.levelno >= logging.DEBUG:
            record.level_char = "*"
        else:
            record.level_char = " "
        return True


class MPRotatingFileHandler(logging.handlers.RotatingFileHandler):
    """
    Handle the uncommon case of the log attempting to roll over when
    another process has the log open. This only happens on Windows, and
    the log ends up being a handful of KBs greater than 1024. Entries
    are still written, and the rollover happens if/when the MainProcess is
    the only process with the log file open.
    """

    def __init__(self, filename, **kwargs):
        # Expand and variables and home directories and make path if it doesn't exist.
        filename = os.path.expandvars(os.path.expanduser(filename))

        # If path is relative, add to standard log directory.
        if not os.path.isabs(filename):
            filename = os.path.join(appdirs.user_log_dir("mwcp", appauthor=False), filename)

        directory = os.path.dirname(filename)
        if not os.path.exists(directory):
            os.makedirs(directory)
        super(MPRotatingFileHandler, self).__init__(filename, **kwargs)

    def doRollover(self):
        """
        Attempt to roll over to the next log file. If the current file
        is locked (Windows issue), keep writing to the original file until
        it is unlocked.

        :return:
        """
        try:
            super(MPRotatingFileHandler, self).doRollover()
        except OSError as e:
            if not (sys.platform == "win32" and e.errno == errno.EACCES):
                raise


class MPChildHandler(logging.Handler):
    """
    Simple handler for child processes.

    Ensures pickle-ability and sends the record entry to the queue.
    """

    def __init__(self, log_queue):
        super(MPChildHandler, self).__init__()
        self.queue = log_queue

    def emit(self, record):
        if record.exc_info:
            record.exc_text = "".join(traceback.format_exception(*record.exc_info))
            record.exc_info = None

        self.queue.put(record)


class ListHandler(logging.Handler):
    """
    Log to a list, with an optional maximum number of records to store.

    Full records are available with the `records` property, and messages (i.e.
    the text of the log entry) at available with the `messages` property.
    """

    def __init__(self, entries=None):
        """
        Behaves essentially identical to any other handler.

        The only option is max_entries, to specify the max number of log
        entries kept. By default, no limit.

        :param int entries: Maximum number of records to store.
        """
        super(ListHandler, self).__init__()

        self._deque = deque(maxlen=entries)

    def __copy__(self):
        new_handler = ListHandler()
        # Actually copy the deque, otherwise we'll get double entries
        new_handler._deque = copy.copy(self._deque)
        return new_handler

    def emit(self, record):
        msg = self.format(record)
        record.formatted_msg = msg
        self._deque.append(record)

    def clear(self):
        return self._deque.clear()

    @property
    def records(self):
        """
        List of the last `max_entries` records logged.
        """
        return list(self._deque)

    @property
    def messages(self):
        """
        List of the last `max_entries` formatted messages logged.
        """
        return [record.formatted_msg for record in self._deque]


def start_listener():
    """Start the listener thread for multi-process logging."""
    if mp.current_process().name != "MainProcess":
        return

    def _mp_log_listener(log_queue):
        while True:
            record = log_queue.get()
            _logger = logging.getLogger(record.name)
            if _logger.isEnabledFor(record.levelno):
                _logger.handle(record)

    listener_thread = threading.Thread(target=_mp_log_listener, args=(mp_queue,))
    listener_thread.daemon = True
    listener_thread.start()


def setup_logging(default_level=logging.INFO, queue=None):
    """
    Sets up logging using default log config file or log config file set by 'MWCP_LOG_CFG'

    :param default_level: Default log level to set to if config file fails.
    :param queue: Queue used to pass logs to.
    """
    if queue:
        assert mp.current_process().name != "MainProcess"
        logging.root.addHandler(MPChildHandler(queue))
        # If on Windows, allow all records to pass through, this is necessary because Windows
        # subprocesses don't duplicate the global state like posix systems.
        # Therefore, we have to pass all log messages through since effective
        # log level is unknown.
        if "Windows" in platform.platform():
            logging.root.setLevel(logging.DEBUG)
    else:
        # Allow setting log configuration using 'MWCP_LOG_CFG' environment variable.
        log_config = os.getenv("MWCP_LOG_CFG", None)
        if log_config is None:
            log_config = mwcp.config.get("LOG_CONFIG_PATH", None)
        else:
            warnings.warn(
                "Using MWCP_LOG_CFG to set log configuration is deprecated. "
                "Please specify path in the configuration file instead."
            )
        if log_config:
            try:
                with open(log_config, "rt") as f:
                    config = yaml.safe_load(f.read())
                logging.config.dictConfig(config)
            except IOError as e:
                warnings.warn("Unable to set log config file: {} with error: {}".format(log_config, e))
                logging.basicConfig(level=default_level)
        else:
            logging.basicConfig(level=default_level)

        # Startup queue listener if we are in the main process.
        start_listener()



"""
Interface for registering and accessing parsers.
"""
import imp
from collections import namedtuple
import hashlib
import importlib
import importlib.machinery
import importlib.util
import logging
import os
import pkgutil
from typing import Optional, Tuple

import pkg_resources

from ruamel.yaml import YAML

yaml = YAML()

from mwcp.parser import Parser
from mwcp.dispatcher import Dispatcher

logger = logging.getLogger(__name__)

ParserInfo = namedtuple("ParserInfo", ("name", "source", "author", "description"))


class Source:
    """Represents a source of parsers."""

    def __init__(self, name, path, config, is_pkg):
        self.name = name
        self.path = path
        self.config = config
        self.is_pkg = is_pkg
        self._package = None

    def __getstate__(self):
        # Modifying __getstate__ in order to null out _package, since that is not picklable.
        _dict = self.__dict__
        _dict["_package"] = None
        return _dict

    @property
    def package(self):
        """Imports and returns the Python package for the source."""
        if not self._package:
            # Import source on first call.
            if self.is_pkg:
                try:
                    self._package = importlib.import_module(self.path)
                except ImportError:
                    raise ValueError("Could not import source: {}".format(self.path))
            else:
                self._package = _create_package(self.path)
        return self._package


class ParserNotFoundError(Exception):
    """This exception gets thrown if a parser can't be found."""


# Set of parser sources mapped to configuration files.
# Sources can be a file path or import path to a python package containing parsers.
# Maps source -> Source object
_sources = {}

_default_source = None


def set_default_source(source_name):
    """
    Sets a default parser source to use if not explicitly defined.
    If this is not set, all sources will be considered.

    :param source_name: The name of the source to set.
    """
    global _default_source
    _default_source = source_name


def clear_default_source():
    """
    Clears a previously set default source.
    """
    global _default_source
    _default_source = None


def clear():
    """
    Removes all registered parsers.
    """
    global _sources
    global _default_source
    _sources = {}
    _default_source = None


def register_entry_points():
    """
    Registers parsers found in entry_point: "mwcp.parsers"
    :return:
    """
    for entry in pkg_resources.iter_entry_points("mwcp.parsers"):
        package = entry.load()
        register_parser_package(package, source_name=entry.name)


def _load_config(config_file_path):
    """
    Loads and validates given parser config file path.

    :raises ValueError: If loaded config file is invalid.
    """
    with open(config_file_path, "r") as fp:
        config = yaml.load(fp)

    logger.debug("Validating parser config: {}".format(config_file_path))
    if not isinstance(config, dict):
        raise ValueError("Parser config is not a dictionary: {}".format(config_file_path))
    config = {str(key): value for key, value in config.items()}  # Force keys to be strings.
    for key, value in config.items():
        if "." in key:
            raise ValueError('"." in group name is not allowed: {}'.format(key))
        if "description" not in value:
            raise ValueError('Missing "description" field in group: {}'.format(key))
        if "parsers" not in value:
            raise ValueError('Missing "parsers" field in group: {}'.format(key))
        if not isinstance(value["parsers"], list):
            raise ValueError('"parsers" field is not a list in group: {}'.format(key))
    return config


def register_parser_directory(directory, config_file_path=None, source_name=None):
    """
    Registers parsers found in parser_dir. This function allows you to register one-off parsers
    that are not part of an installed python package.

    :param str directory: An extra directory to look for one-off parsers.
    :param config_file_path: Optional path to a parser configuration file used to define parser groups.
            If not provided, it will attempt to pull from the "config" attribute of the __init__ module.
    :param source_name: Unique name to give to the source. (uses directory path otherwise)

    :raises ValueError: If loaded config file is invalid.
    """
    global _sources

    if not os.path.isdir(directory):
        raise ValueError(u"Parser directory not found or not a directory: {}".format(directory))

    # Ensure this directory can be converted to a package and pull config_file_path if available.
    package = _create_package(directory)
    if not config_file_path:
        config_file_path = getattr(package, "config", None)
    config = _load_config(config_file_path) if config_file_path else {}

    if not source_name:
        source_name = directory

    # NOTE: _sources must be discoverable without modification, so we can't register the package.
    _sources[source_name] = Source(source_name, directory, config, False)


def register_parser_package(package, config_file_path=None, source_name=None):
    """
    Registers Python package containing MWCP parsers.

    :param package: An Python package containing submodules that contain MWCP parsers.
        NOTE: Package must be discoverable in subprocesses without modifying the python path.
              Please use register_parser_directory() instead if that is not possible.
    :param config_file_path: Path to parser configuration file used to define parser groups.
        If not provided, it will attempt to pull from the "config" attribute of the __init__ module.
    :param source_name: Unique name to give to the source. (uses package name otherwise)

    :raises AttributeError: If config_file_path is not provided and package doesn't have a "config" attribute.
    :raises ValueError: If loaded config file is invalid.
    """
    global _sources

    if not hasattr(package, "__path__"):
        raise ValueError("{!r} is not a Python package".format(package))

    if not config_file_path:
        config_file_path = getattr(package, "config", None)
    config = _load_config(config_file_path) if config_file_path else {}

    if not source_name:
        source_name = package.__name__.lower()

    _sources[source_name] = Source(source_name, package.__name__, config, True)


def _create_package(directory):
    """Creates a Python package object from given directory."""
    # Create a dummy package for the directory.
    package_name = hashlib.md5(directory.encode("utf8")).hexdigest()
    package_init = os.path.join(directory, "__init__.py")
    # Create __init__.py if it doesn't exist.
    if not os.path.exists(package_init):
        logger.info("Creating required __init__ module: {}".format(package_init))
        with open(package_init, "a"):
            pass
    try:
        package = imp.load_source(package_name, package_init)
        package.__path__ = [directory]
    except IOError as e:
        raise ValueError("Could not create package from {} with error: {}".format(directory, e))
    return package


def _is_module_available(module_name):
    """Determines whether given module name is available without importing."""
    try:
        return bool(importlib.util.find_spec(module_name))
    except ModuleNotFoundError:
        # If we get this error, that means an intermediate package couldn't be found.
        return False


def _generate_parser_aux(parser_name, group_name, source, _visiting):
    """
    Auxiliary function used by _generate_parser() to format the parser_name
    before running _generate_parser()
    """
    orig_parser_name = parser_name

    if parser_name.startswith("."):
        parser_name = group_name + parser_name

    # Pull out imported source.
    if ":" in parser_name:
        source_name, _, parser_name = parser_name.partition(":")
        if source_name not in _sources:
            raise RuntimeError(f"Unable to find source: {source_name}")
        source = _sources[source_name]

    if (parser_name, source.name) in _visiting:
        raise RuntimeError(f"Detected recursive loop: {group_name} -> {orig_parser_name}")

    try:
        return _generate_parser(parser_name, source, _visiting=_visiting)
    except ParserNotFoundError as e:
        raise RuntimeError(f"Unable to find {parser_name} with error: {e}")


def _generate_parser(name: str, source: Source, recursive=True, _visiting=None):
    """
    Generates parser for given name.

    :param str name: Name of parser or parser group.
    :param Source source: Source object containing parser.
    :param bool recursive: Recursively generate listed sub parsers.
        (otherwise only top level parsers will be produced)
    :param _visiting: Used internally for recursive loop detection.

    :returns: Either a Dispatcher object for a group of parsers or a Parser class.

    :raises ParserNotFound: If parser could not be found.
    """
    if _visiting is None:
        _visiting = set()

    _visiting.add((name, source.name))

    try:
        # First check if parser name is a parser group.
        options = dict(source.config[name])
    except KeyError:
        logger.debug("Generating parser: {}".format(name))
        if "." not in name:
            raise ParserNotFoundError("Invalid name {}".format(name))
        # If not, find and import the referenced mwcp.Parser class.
        module_name, _, class_name = name.rpartition(".")
        module_fullname = source.package.__name__ + "." + module_name

        logger.debug("Checking existence of {}".format(module_fullname))
        if not _is_module_available(module_fullname):
            raise ParserNotFoundError("{} module does not exist".format(module_fullname))

        logger.debug("Importing: {}".format(module_fullname))
        module = importlib.import_module(module_fullname)

        if not hasattr(module, class_name):
            raise ParserNotFoundError("{} is not in {}".format(class_name, module_fullname))

        klass = getattr(module, class_name)

        if not issubclass(klass, Parser):
            raise ParserNotFoundError("{}.{} is not a mwcp.Parser class".format(module_fullname, class_name))

        klass.name = name
        klass.source = source.name
        logger.debug("Created parser: {!r}".format(klass))
        _visiting.remove((name, source.name))
        return klass

    # Otherwise, instantiate a mwcp.Dispatcher class for the parser group.
    group_name = name
    parser_names = options.pop("parsers")
    sub_parsers = []
    if recursive:
        for parser_name in parser_names:
            sub_parsers.append(_generate_parser_aux(parser_name, group_name, source, _visiting))

    # Dereference default parser.
    default = options.pop("default", None)
    if default and recursive:
        options["default"] = _generate_parser_aux(default, group_name, source, _visiting)

    parser = Dispatcher(group_name, source.name, parsers=sub_parsers, **options)
    logger.debug("Created parser group: {!r}".format(parser))
    _visiting.remove((name, source.name))
    return parser


def _import_all_modules(package):
    """Recursively imports all modules from a given python package or directory."""
    for _, name, is_pkg in pkgutil.walk_packages(package.__path__):
        full_name = "{}.{}".format(package.__name__, name)
        module = importlib.import_module(full_name)
        if is_pkg:
            _import_all_modules(module)


def iter_parsers(name: str = None, source: str = None, config_only=True, _recursive=True):
    """
    Iterates all registered parsers.

    :param str name: Filters parser based on a particular name. (":" notation is also supported)
    :param str source: Filters parser based on a particular source.
                       (source is either the name of a python package or path to local directory)
    :param bool config_only: Whether to only include parsers listed in the parser configuration file.
                             (ie. ignore component parsers like "Foo.Implant")
    :param bool _recursive: Whether to generate sub parsers.
        (This is used internally, don't change it unless you know what you are doing)

    :yields: tuple containing: (Source tuple, parser)

    :raises ValueError: If a parser name or source could not be found.
    """
    global _sources

    if name and not source:
        # If name is using ":" notation, assume it is being organized by "source_name:parser_name"
        # (os.path.basename is necessary in-case source is a file path containing ":"'s)
        orig_name = name
        _, _, name = os.path.basename(name).rpartition(":")
        source = orig_name[: -(len(name) + 1)]

    # Use default source if one is not provided.
    source = source or _default_source or None

    sources = []
    if source:
        if source in _sources:
            sources.append((source, _sources[source]))
    else:
        sources += _sources.items()

    for source_name, source in sources:
        # Find list of parser names to generate
        if name:
            try:
                parser = _generate_parser(name, source, recursive=_recursive)
                yield source, parser
            except ParserNotFoundError as e:
                logger.debug("[{}] {}".format(source_name, e))
                # Parser couldn't be found for this source.
                continue
        else:
            # If parser name is not provided provide all parsers from the given source.
            for parser_name in source.config.keys():
                parser = _generate_parser(parser_name, source, recursive=_recursive)
                yield source, parser

            # Also list all the component parsers if requested.
            if not config_only:
                _import_all_modules(source.package)
                package_prefix = source.package.__name__ + "."
                for klass in set(Parser.iter_subclasses()):
                    # Ignore classes without DESCRIPTIONS since they are usually base classes.
                    if klass.DESCRIPTION and klass.__module__.startswith(package_prefix):
                        parser_name = "{}.{}".format(klass.__module__[len(package_prefix):], klass.__name__)
                        klass.name = parser_name
                        yield source, klass


def get_parser_descriptions(name=None, source=None, config_only=True):
    """
    Retrieve list of parser descriptions

    :param str name: Filters parser based on a particular name. (":" notation is also supported)
    :param str source: Filters parser based on a particular source.
                       (source is either the name of a python package or path to local directory)
    :param bool config_only: Whether to only include parsers listed in the parser configuration file.
                             (ie. ignore component parsers like "Foo.Implant")

    Returns list of tuples per parser. Tuple contains parser name, author, and description.
    """
    descriptions = []
    for _source, parser in iter_parsers(name=name, source=source, config_only=config_only, _recursive=False):
        descriptions.append(ParserInfo(parser.name, _source.name, parser.AUTHOR, parser.DESCRIPTION))
    return sorted(descriptions, key=lambda e: tuple(sub.lower() for sub in e))  # Case-insensitive sorting.



"""
This is an example parser used to show how to use the Dispatcher model to organize Parsers by their components.
"""

import re

from mwcp import Parser, FileObject
from mwcp.utils import construct
from mwcp.utils import pefileutils


class Carrier(Parser):
    """A carrier for the Bar dropper."""

    DESCRIPTION = "Bar Carrier"

    MARKER = re.compile(b"BARCAR\x01\x02(?P<embed_data>.+)\x02\x01CARBAR")

    @classmethod
    def identify(cls, file_object):
        """
        Identifies Bar carrier by looking for a PE file that has the BARCAR\x01\x02 carrier.

        :returns: boolean indicating if file is a Bar Carrier.
        """
        return file_object.pe and cls.MARKER.search(file_object.file_data)

    def run(self):
        """Extracts config and embedded Bar dropper."""
        # Pull embedded dropper.
        match = self.MARKER.search(self.file_object.file_data)
        embed_data = match.group("embed_data")

        # Dispatch embedded dropper to be picked up by another ComponentParser.
        self.dispatcher.add_to_queue(FileObject(embed_data, self.reporter))


class Dropper(Parser):
    """A dropper for the Bar implant."""

    DESCRIPTION = "Bar Dropper"

    @classmethod
    def identify(cls, file_object):
        """
        Identifies Bar dropper by looking for '.DROPPER' resource directory.

        :returns: boolean indicating if file is a Bar Dropper.
        """
        return file_object.pe and pefileutils.check_rsrc_dir(".DROPPER", pe=file_object.pe)

    def run(self):
        """Extracts config and embedded Implant files."""
        # Dispatch all the files in the ".DROPPER" resource directory.
        for rsrc in pefileutils.iter_rsrc(self.file_object.pe, ".DROPPER"):
            self.dispatcher.add_to_queue(FileObject(rsrc.data, self.reporter, file_name=rsrc.idname))


class Implant(Parser):
    """A Bar implant found as a python script."""

    DESCRIPTION = "Bar Implant"

    @classmethod
    def identify(cls, file_object):
        """
        Identifies Bar implant by looking for the .py extension.
        (This condition is unique enough for the Bar family for now.)

        :returns: boolean indicating if file is a Bar Dropper.
        """
        return file_object.file_name.endswith(".py")

    def run(self):
        """Extracts config and embedded Implant files."""
        # Extract the callback and port numbers found in this python script.
        for match in re.finditer(b"http://(?P<callback>.+?):(?P<port>\d+)", self.file_object.file_data):
            callback, port = match.groups()
            self.reporter.add_metadata("c2_socketaddress", (callback, port, "tcp"))


class ImplantV2(Parser):
    """An alternative version of Bar Implant using an exe instead."""

    DESCRIPTION = "Bar Implant V2"

    CALLBACK_RE = re.compile(b"BARCALL:(?P<callback>.+?):(?P<port>\d+)", re.DOTALL)
    CALLBACK = construct.Regex(CALLBACK_RE, port=construct.Int32ul)

    @classmethod
    def identify(cls, file_object):
        """
        Identifies Bar implant by looking for a PE file contains some markers we are looking for.
        (Since this ComponentParser is the last in the list

        :returns: boolean indicating if file is a Bar Dropper.
        """
        return file_object.pe and cls.CALLBACK_RE.search(file_object.file_data)

    def run(self):
        """Extracts config and embedded Implant files."""
        # Extract the callback and port numbers found in this .exe script.
        for config in self.CALLBACK[:].parse(self.file_object.file_data):
            self.reporter.add_metadata("c2_socketaddress", (config.callback, str(config.port), "tcp"))



"""
Wrapper for Techanarchy RATDecoders using techanarchy_bridge
"""

import os
from pathlib import Path

from mwcp import Parser
from mwcp.resources import RATDecoders

RAT_DECODERS = [decoder.stem for decoder in Path(RATDecoders.__file__).parent.glob("[!_]*.py")]


def run(self):
    from mwcp.resources import techanarchy_bridge

    name = self.__class__.__name__
    scriptpath = os.path.join(os.path.dirname(RATDecoders.__file__), name + ".py")
    techanarchy_bridge.run_decoder(self, scriptpath)


# Dynamically declare Parser classes.
for name in RAT_DECODERS:
    if name == "TEMPLATE":
        continue
    klass = type(name, (Parser,), {"DESCRIPTION": name, "run": run, "AUTHOR": "TechAnarchy"})
    klass.__module__ = __name__  # Module originally gets incorrectly set to "abc"
    globals()[name] = klass



"""
DC3-MWCP Framework command line tool.

Used for running and testing parsers.
"""

from __future__ import print_function, division
from future.builtins import str, zip

import csv
import datetime
import glob
import hashlib
from io import open
import itertools
import json
import logging
import os
import shutil
import subprocess
import sys
import timeit
import traceback
import warnings

import click
import six
import tabulate

import mwcp
from mwcp.tester import Tester
from mwcp.utils.stringutils import convert_to_unicode


logger = logging.getLogger("mwcp")


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.option("-d", "--debug", is_flag=True, help="Enables DEBUG level logs.")
@click.option("-v", "--verbose", is_flag=True, help="Enables INFO level logs.")
@click.option(
    "-c",
    "--config",
    "config_path",
    type=click.Path(exists=True, dir_okay=False),
    help="File path to configuration file.",
    default=mwcp.config.user_path,
    show_default=True,
    envvar="MWCP_CONFIG",
    show_envvar=True,
)
@click.option(
    "--parser-dir",
    type=click.Path(exists=True, file_okay=False),
    help="Optional extra parser directory.",
)
@click.option(
    "--parser-config",
    type=click.Path(exists=True, dir_okay=False),
    help="Optional parser configuration file to use with extra parser directory.",
)
@click.option(
    "--parser-source",
    help="Set a default parsers source to use. If not provided parsers from all sources will be available.",
)
def main(debug, verbose, config_path, parser_dir, parser_config, parser_source):
    # Setup configuration
    mwcp.config.load(config_path)
    if parser_dir:
        mwcp.config["PARSER_DIR"] = parser_dir
    parser_dir = mwcp.config.get("PARSER_DIR")
    if parser_config:
        mwcp.config["PARSER_CONFIG_PATH"] = parser_config
    parser_config = mwcp.config.get("PARSER_CONFIG_PATH")
    if parser_source:
        mwcp.config["PARSER_SOURCE"] = parser_source
    parser_source = mwcp.config.get("PARSER_SOURCE")

    # Setup logging
    mwcp.setup_logging()
    if debug:
        logging.root.setLevel(logging.DEBUG)
    elif verbose:
        logging.root.setLevel(logging.INFO)
    # else let log_config.yaml set log level.

    # Register parsers
    mwcp.register_entry_points()
    if parser_dir:
        mwcp.register_parser_directory(parser_dir, config_file_path=parser_config)
    if parser_source:
        mwcp.set_default_source(parser_source)


@main.command()
@click.option("--host", default="127.0.0.1", show_default=True, help="The interface to bind to.")
@click.option("--port", default=8080, show_default=True, help="The port to bind to.")
@click.option("--debug", is_flag=True, help="Show the interactive debugger if errors occur.")
def serve(host, port, debug):
    """Run a server to handle parsing requests."""
    from mwcp.tools import server

    if debug:
        os.environ["FLASK_ENV"] = "development"

    app = server.create_app()
    app.run(host=host, port=port, debug=debug, use_reloader=False)


@main.command()
def config():
    """Opens up configuration file for editing."""
    file_path = mwcp.config.user_path
    if sys.platform == "win32":
        try:
            os.startfile(file_path, "edit")
        except WindowsError:
            os.startfile(file_path)
    else:
        opener = "open" if sys.platform == "darwin" else "xdg-open"
        subprocess.call([opener, file_path])

    # TODO: Add a "-u" flag to this command when we need to update the user's configuration file
    # with new fields.


@main.command("list")
@click.option(
    "-a",
    "--all",
    "all_",
    is_flag=True,
    help="Whether to also include parsers not listed in any parsers configuration file.",
)
@click.option("-j", "--json", "json_", is_flag=True, help="Display as JSON output.")
def list_(all_, json_):
    """Lists registered malware config parsers."""
    descriptions = mwcp.get_parser_descriptions(config_only=not all_)
    if json_:
        print(json.dumps(descriptions, indent=4))
    else:
        print(tabulate.tabulate(descriptions, headers=["NAME", "SOURCE", "AUTHOR", "DESCRIPTION"]))


# Column order for standard fields in csv.
# (All other fields will be appended in alphabetical order after these)
_STD_CSV_COLUMNS = [
    "scan_date",
    "inputfilename",
    "outputfile.name",
    "outputfile.description",
    "outputfile.md5",
    "outputfile.base64",
]


def _format_metadata_value(v):
    """Formats metadata value to a human readable unicode string."""
    if isinstance(v, (list, tuple)):
        result = u""
        for j in v:
            if not isinstance(j, (bytes, str)):
                result += u"{}\n".format(u", ".join(map(convert_to_unicode, j)))
            else:
                result += u"{}\n".format(convert_to_unicode(j))
        return result.rstrip()
    elif isinstance(v, dict):
        result = u""
        for field, value in six.iteritems(v):
            if isinstance(value, (list, tuple)):
                value = u"[{}]".format(u", ".join(value))

            result += u"{}: {}\n".format(field, value)
        return result.rstrip()
    else:
        return convert_to_unicode(v)


def _write_csv(input_files, results, csv_path=None):
    """
    Writes out results as a csv.

    :param input_files: List of filenames for each respective metadata.
    :param results: List of metadata dictionaries.
    :param csv_path: Path to write out csv file, defaults to stdout.

    :raises IOError: If csv could not be written out.
    """
    scan_date = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    # Add/Teak metadata.
    for inputfilename, metadata in zip(input_files, results):
        # Add scan date.
        metadata[u"scan_date"] = scan_date
        if u"inputfilename" not in metadata:
            metadata[u"inputfilename"] = inputfilename

        # Flatten 'other' entry so nested values get their own columns,
        # are more readable, and easier to individually analyze.
        #
        # Example:
        #   {'other': {"unique_entry": "value", "unique_key": "value2"}}
        #   Results in columns: other, other.unique_entry, other.unique_key
        if u"other" in metadata:
            for sub_key, sub_value in metadata[u"other"].items():
                metadata[u"other.{}".format(convert_to_unicode(sub_key))] = sub_value
            del metadata[u"other"]

        # Split outputfile into multiple fields.
        if u"outputfile" in metadata:
            value = list(zip(*metadata[u"outputfile"]))
            metadata[u"outputfile.name"] = value[0]
            metadata[u"outputfile.description"] = value[1]
            metadata[u"outputfile.md5"] = value[2]
            del metadata[u"outputfile"]

    # Sort columns, but with PREFIX_COLUMNS showing up first.
    column_names = set(itertools.chain(*(metadata.keys() for metadata in results)))
    column_names = sorted(
        column_names, key=lambda x: str(_STD_CSV_COLUMNS.index(x)) if x in _STD_CSV_COLUMNS else str(x)
    )

    # Reformat metadata and write to CSV
    if csv_path is None:
        csvfile = sys.stdout
    else:
        csvfile = open(csv_path, "wb" if six.PY2 else "w")

    try:
        dw = csv.DictWriter(csvfile, fieldnames=column_names, lineterminator="\n")
        dw.writeheader()
        for metadata in results:
            dw.writerow({k: _format_metadata_value(v) for k, v in metadata.items()})
    finally:
        if csv_path:
            csvfile.close()


def _parse_file(reporter, file_path, parser, include_filename=False):
    """
    Parses given file_path with given parser.

    :param reporter: Reporter to use for parsing.
    :param file_path: File path to parse or "-" for stdin
    :param parser: Name of parser to run (can use ":" notation)
    :param include_filename: Whether to include input file metadata in the results.
    :return: Dictionary of results.
    """
    if file_path == "-":
        reporter.run_parser(parser, data=sys.stdin.read())
    else:
        reporter.run_parser(parser, file_path)

    result = reporter.metadata
    input_file = reporter.input_file

    # TODO: This should just be included by the reporter by default.
    if include_filename:
        result["inputfilename"] = file_path
        result["md5"] = input_file.md5
        result["sha1"] = input_file.sha1
        result["sha256"] = input_file.sha256
        result["parser"] = parser
        if input_file.pe:
            result["compiletime"] = datetime.datetime.fromtimestamp(
                reporter.input_file.pe.FILE_HEADER.TimeDateStamp
            ).isoformat()

    if reporter.errors:
        result["errors"] = reporter.errors

    return result


@main.command()
@click.option("-f", "--format", type=click.Choice(["csv", "json"]), help="Displays results in another format.")
@click.option(
    "-o",
    "--output-dir",
    type=click.Path(exists=True, file_okay=False),
    help="Root output directory to store residual files. (defaults to current directory)",
)
@click.option(
    "--output-files/--no-output-files", default=True, show_default=True, help="Whether to output files to filesystem."
)
@click.option(
    "--cleanup/--no-cleanup", default=True, show_default=True, help="Whether to cleanup temporary files after parsing."
)
@click.option(
    "--prefix/--no-prefix", default=True, show_default=True,
    help="Whether to prefix output filenames with the first 5 characters of the md5. "
         "If turned off, unique files with the same file name will be overwritten."
)
@click.option(
    "-i",
    "--include-filename",
    is_flag=True,
    help="Include file information such as filename, hashes, and compile time in parser output.",
)
@click.argument("parser", required=True)
@click.argument("input", nargs=-1, type=click.Path())
def parse(parser, input, format, output_dir, output_files, cleanup, prefix, include_filename):
    """
    Parses given input with given parser.

    \b
    PARSER: Name of parser to run.
    INPUT: One or more input file paths. (Wildcards are allowed).

    \b
    Common usages::
        mwcp parse foo ./malware.bin                          - Run foo parser on ./malware.bin
        mwcp parse foo ./repo/*                               - Run foo parser on files found in repo directory.
        mwcp parse -f json foo ./malware.bin                  - Run foo parser and display results as json.
        mwcp parse -f csv foo ./repo/* > ./results.csv        - Run foo parser on a directory and output results as a csv file.
    """
    # Python won't process wildcards when used through Windows command prompt.
    if any("*" in path for path in input):
        new_input = []
        for path in input:
            if "*" in path:
                new_input.extend(glob.glob(path))
            else:
                new_input.append(path)
        input = new_input

    input_files = list(filter(os.path.isfile, input))
    output_dir = output_dir or ""

    # Run MWCP
    try:
        results = []
        for path in input_files:
            reporter = mwcp.Reporter(
                # Store output files to a folder with the same name as the input file.
                outputdir=os.path.join(output_dir, os.path.basename(path) + "_mwcp_output"),
                disable_output_files=not output_files,
                disable_temp_cleanup=not cleanup,
                prefix_output_files=prefix,
            )
            logger.info("Parsing: {}".format(path))
            result = _parse_file(reporter, path, parser, include_filename=include_filename)
            results.append(result)
            if not format:
                reporter.print_report()

        if format == "csv":
            _write_csv(input_files, results)
        elif format == "json":
            print(json.dumps(results, indent=4))

    except Exception as e:
        error_message = "Error running DC3-MWCP: {}".format(e)
        traceback.print_exc()
        if format == "json":
            print(json.dumps({"errors": [error_message]}))
        else:
            print(error_message)
        sys.exit(1)


def _get_malware_repo_path(file_path):
    """
    Gets file path for a file in the malware_repo based on the md5 of the given file_path.
    """
    if not mwcp.config.get("MALWARE_REPO"):
        raise ValueError("Malware Repository not set.")
    with open(file_path, "rb") as fo:
        md5 = hashlib.md5(fo.read()).hexdigest()
    return os.path.join(mwcp.config["MALWARE_REPO"], md5[0:4], md5)


def _add_to_malware_repo(file_path):
    """
    Adds the given file path to the malware repo.
    Returns resulting destination path.
    """
    dest_path = _get_malware_repo_path(file_path)
    dest_dir = os.path.dirname(dest_path)

    if os.path.isfile(dest_path):
        click.echo("File already exists in malware repo: {}".format(dest_path))
        return dest_path

    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)

    click.echo("Copying {} to {}".format(file_path, dest_path))
    shutil.copy(file_path, dest_path)
    return dest_path


def _read_input_list(filename):
    inputfilelist = []
    if filename:
        if filename == "-":
            inputfilelist = [line.rstrip() for line in sys.stdin]
        else:
            with open(filename, "rb") as f:
                inputfilelist = [line.rstrip() for line in f]

    return inputfilelist


def _median(data):
    """
    'Borrowed' from Py3's statistics library.

    :param data: Data to get median of
    :return: Median as a float or int
    :rtype: float or int
    """
    data = sorted(data)
    length = len(data)
    if length == 0:
        raise ValueError("No median for empty data.")
    elif length % 2 == 1:
        return data[length // 2]
    else:
        i = length // 2
        return (data[i - 1] + data[i]) / 2


def _run_tests(tester, silent=False, show_passed=False):
    print("Running test cases. May take a while...")

    start_time = timeit.default_timer()
    test_results = []
    all_passed = True
    total = tester.total
    failed = []

    # Generate format string.
    digits = len(str(total))
    if not tester.test_cases:
        parser_len = 10
        filename_len = 10
    else:
        parser_len = max(len(test_case.parser) for test_case in tester.test_cases)
        filename_len = max(len(test_case.filename) for test_case in tester.test_cases)
    msg_format = "{{parser:{0}}} {{filename:{1}}} {{run_time:.4f}}s".format(parser_len, filename_len)

    format_str = "{{count:> {0}d}}/{{total:0{0}d}} - ".format(digits) + msg_format

    # Run tests and output progress results.
    for count, test_result in enumerate(tester, start=1):
        all_passed &= test_result.passed
        if not test_result.passed:
            failed.append((count, test_result.parser, test_result.filename))

        if test_result.run_time:  # Ignore missing tests from stat summary.
            test_results.append(test_result)

        if not silent:
            message = format_str.format(
                count=count,
                total=total,
                parser=test_result.parser,
                filename=test_result.filename,
                run_time=test_result.run_time,
            )
            # Skip print() to immediately flush stdout buffer (issue in Docker containers)
            sys.stdout.write(message + "\n")
            sys.stdout.flush()
            test_result.print(failed_tests=True, passed_tests=show_passed)

    end_time = timeit.default_timer()

    # Present test statistics
    if not silent and test_results:
        print("\nTest stats:")
        print("\nTop 10 Slowest Test Cases:")

        format_str = "{index:2}. " + msg_format

        # Cases sorted slowest first
        sorted_cases = sorted(test_results, key=lambda x: x.run_time, reverse=True)
        for i, test_result in enumerate(sorted_cases[:10], start=1):
            print(
                format_str.format(
                    index=i, parser=test_result.parser, filename=test_result.filename, run_time=test_result.run_time
                )
            )

        print("\nTop 10 Fastest Test Cases:")
        for i, test_result in enumerate(list(reversed(sorted_cases))[:10], start=1):
            print(
                format_str.format(
                    index=i, parser=test_result.parser, filename=test_result.filename, run_time=test_result.run_time
                )
            )

        run_times = [test_result.run_time for test_result in test_results]
        print("\nMean Running Time: {:.4f}s".format(sum(run_times) / len(test_results)))
        print("Median Running Time: {:.4f}s".format(_median(run_times)))
        print("Cumulative Running Time: {}".format(datetime.timedelta(seconds=sum(run_times))))
        print()

    print("Total Running Time: {}".format(datetime.timedelta(seconds=end_time - start_time)))

    if failed:
        print()
        print("Failed tests:")
        for test_info in failed:
            print("#{} - {}\t{}".format(*test_info))
        print()

    print("All Passed = {0}\n".format(all_passed))
    exit(0 if all_passed else 1)


@main.command()
@click.option(
    "-t",
    "--testcase-dir",
    type=click.Path(file_okay=False),
    help="Directory containing JSON test case files. (defaults to a "
    '"tests" directory located within the root of the parsers directory)',
)
@click.option(
    "-m",
    "--malware-repo",
    type=click.Path(file_okay=False),
    help="Directory containing malware samples used for testing.",
)
# Arguments used for run test cases.
@click.option(
    "-n", "--nprocs", type=int, help="Number of test cases to run simultaneously. [default: 3/4 * logical CPU cores]"
)
# Arguments used to generate and update test cases
@click.option(
    "-u",
    "--update",
    is_flag=True,
    help="Update all stored test cases with newly produced results. "
    "If used with the --add option, this allows the test cases for the added files to "
    "be updated if the file already exists in the test case.",
)
@click.option(
    "-a",
    "--add",
    multiple=True,
    type=click.Path(exists=True, dir_okay=False),
    help="Adds given file to the test case. (Will first copy file to malware repo if provided.)",
)
@click.option(
    "-i",
    "--add-filelist",
    multiple=True,
    type=click.Path(exists=True, dir_okay=False),
    help="Adds a file of file paths to the test case.",
)
@click.option(
    "-x",
    "--delete",
    multiple=True,
    type=click.Path(exists=True, dir_okay=False),
    help="Deletes given file from the test case. " "(Note, this does not delete the file if placed in a malware repo.)",
)
@click.option("-y", "--yes", is_flag=True, help="Auto confirm questions.")
@click.option("--force", is_flag=True, help="Force test case add/update when errors are encountered.")
# Arguments to configure console output
@click.option(
    "-f",
    "--show-passed",
    is_flag=True,
    help="Display tests case details for passed tests as well." "By default only failed tests are shown.",
)
@click.option("-s", "--silent", is_flag=True, help="Limit output to statement saying whether all tests passed or not.")
# Parser to process.
@click.argument("parser", nargs=-1, required=False)
def test(
    testcase_dir, malware_repo, nprocs, update, add, add_filelist, delete, yes, force, show_passed, silent, parser
):
    """
    Testing utility to create and execute parser test cases.

    \b
    PARSER: Parsers to test. Test all parers if not provided.

    \b
    Common usages::
        mwcp test                                             - Run all tests cases.
        mwcp test foo                                         - Run test cases for foo parser.
        mwcp test foo -u                                      - Update existing test cases for foo parser.
        mwcp test -u                                          - Update existing test cases for all parsers.
        mwcp test foo --add=./malware.bin                     - Add test case for malware.bin sample for foo parser.
        mwcp test foo -u --add=./malware.bin                  - Add test case for malware.bin sample.
                                                                Allow updating if a test case for this file already exists.
        mwcp test foo --add-filelist=./paths.txt              - Add tests cases for foo parser using text file of paths.
        mwcp test foo --delete=./malware.bin                  - Delete test case for malware.bin sample for foo parser.
    """
    # Overwrite configuration with command line flags.
    if testcase_dir:
        mwcp.config["TESTCASE_DIR"] = testcase_dir
    if malware_repo:
        mwcp.config["MALWARE_REPO"] = malware_repo

    # Configure test object
    reporter = mwcp.Reporter(disable_output_files=True)
    tester = Tester(reporter=reporter, parser_names=parser or [None], nprocs=nprocs,)

    # Add/Delete
    if add or add_filelist or delete:
        click.echo("Adding new test cases. May take a while...")
        if not parser:
            # Don't allow adding a file to ALL test cases.
            raise click.BadParameter("PARSER must be provided when adding or deleting a file from a test case.")

        # Cast tuple to list so we can manipulate.
        add = list(add)
        for filelist in add_filelist:
            with open(filelist, "r") as f:
                for file_path in f.readlines():
                    add.append(file_path.rstrip("\n"))

        for file_path in add:
            if mwcp.config.get("MALWARE_REPO"):
                file_path = _add_to_malware_repo(file_path)
            tester.add_test(file_path, force=force, update=update)

        for file_path in delete:
            if mwcp.config.get("MALWARE_REPO"):
                file_path = _get_malware_repo_path(file_path)
            tester.remove_test(file_path)

    # Update
    elif update:
        if not parser and not yes:
            click.confirm("WARNING: About to update test cases for ALL parsers. Continue?", abort=True)
        click.echo("Updating test cases. May take a while...")
        tester.update_tests(force=force)

    # Run tests
    else:
        if not parser and not yes:
            click.confirm("PARSER argument not provided. Run tests for ALL parsers?", default=True, abort=True)
        # Force ERROR level logs so we don't spam the console.
        logging.root.setLevel(logging.ERROR)
        _run_tests(tester, silent, show_passed)


if __name__ == "__main__":
    main(sys.argv[1:])



"""
A central location to store common windows enumerations.
This module will be imported along with 'from mwcp.utils import construct'
"""

from .version28 import *

# Visible interface. Add the classes and functions you would like to be available for users of construct
# library here.
__all__ = ['RegHive', 'LanguageIdentifier', 'KnownFolderID', 'AlgorithmID']


REGHIVES = {
    "HKCR": 0x80000000,
    "HKCU": 0x80000001,
    "HKLM": 0x80000002,
    "HKU":  0x80000003,
    "HKPD": 0x80000004,
    "HKCC": 0x80000005,
    "HKDD": 0x80000006,
}


def RegHive(subcon):
    r"""
    Converts an integer to registry hive enum.

    >>> RegHive(Int32ul).build("HKCU")
    '\x01\x00\x00\x80'
    >>> str(RegHive(Int32ul).parse('\x01\x00\x00\x80'))
    'HKCU'
    """
    return Enum(subcon, **REGHIVES)


# TODO: Extend dictionary to incorporate more languages
LANGUAGEIDENTIFIERS = {
    "English (United States)": 0x409,
    "Korean": 0x412,
    "Chinese (PRC)": 0x804,
}


def LanguageIdentifier(subcon):
    r"""
    Converts an integer to language identifer enum

    >>> LanguageIdentifier(Int32ul).build("English (United States)")
    '\t\x04\x00\x00'
    >>> str(LanguageIdentifier(Int32ul).parse("\x04\x08\x00\x00"))
    'Chinese (PRC)'
    """
    return Enum(subcon, **LANGUAGEIDENTIFIERS)


CSIDL = {
    'CSIDL_SYSTEM': 37,
    'CSIDL_COMMON_PROGRAMS': 23,
    'CSIDL_PROFILE': 40,
    'CSIDL_ALTSTARTUP': 29,
    'CSIDL_LOCAL_APPDATA': 28,
    'CSIDL_PRINTHOOD': 27,
    'CSIDL_FONTS': 20,
    'CSIDL_PROGRAM_FILES_COMMON': 43,
    'CSIDL_PROGRAM_FILESX86': 42,
    'CSIDL_MYDOCUMENTS': 5,
    'CSIDL_MYVIDEO': 14,
    'CSIDL_PROGRAM_FILES': 38,
    'CSIDL_ADMINTOOLS': 48,
    'CSIDL_COMMON_DOCUMENTS': 46,
    'CSIDL_CONNECTIONS': 49,
    'CSIDL_COMMON_ALTSTARTUP': 30,
    'CSIDL_DRIVES': 17,
    'CSIDL_RESOURCES_LOCALIZED': 57,
    'CSIDL_HISTORY': 34,
    'CSIDL_NETHOOD': 19,
    'CSIDL_CDBURN_AREA': 59,
    'CSIDL_COMMON_DESKTOPDIRECTORY': 25,
    'CSIDL_SYSTEMX86': 41,
    'CSIDL_COMMON_TEMPLATES': 45,
    'CSIDL_MYPICTURES': 39,
    'CSIDL_COMMON_VIDEO': 55,
    'CSIDL_COMMON_STARTMENU': 22,
    'CSIDL_COMMON_FAVORITES': 31,
    'CSIDL_INTERNET_CACHE': 32,
    'CSIDL_WINDOWS': 36,
    'CSIDL_COMMON_PICTURES': 54,
    'CSIDL_COMMON_APPDATA': 35,
    'CSIDL_DESKTOPDIRECTORY': 16,
    'CSIDL_RESOURCES': 56,
    'CSIDL_COMMON_MUSIC': 53,
    'CSIDL_COMMON_OEM_LINKS': 58,
    'CSIDL_NETWORK': 18,
    'CSIDL_COOKIES': 33,
    'CSIDL_COMPUTERSNEARME': 61,
    'CSIDL_COMMON_ADMINTOOLS': 47,
    'CSIDL_APPDATA': 26,
    'CSIDL_TEMPLATES': 21,
    'CSIDL_COMMON_STARTUP': 24,
    'CSIDL_MYMUSIC': 13,
    'CSIDL_PROGRAM_FILES_COMMONX86': 44
}


def KnownFolderID(subcon):
    r"""
    Converts an integer to a CSIDL (KNownFolderID) value

    >>> KnownFolderID(Int32ul).build("CSIDL_SYSTEM")
    '%\x00\x00\x00'
    >>> str(KnownFolderID(Int32ul).parse("\x18\x00\x00\x00"))
    'CSIDL_COMMON_STARTUP'
    """
    return Enum(subcon, **CSIDL)


ALGIDS = {
    'CALG_DSS_SIGN': 0x00002200,
    'CALG_DES': 0x00006601,
    'CALG_DH_EPHEM': 0x0000aa02,
    'CALG_3DES': 0x00006603,
    'CALG_DESX': 0x00006604,
    'CALG_ECDH': 0x0000aa05,
    'CALG_NO_SIGN': 0x00002000,
    'CALG_DH_SF': 0x0000aa01,
    'CALC_SSL3_SHAMD5': 0x00008008,
    'CALG_3DES_112': 0x00006609,
    'CALG_SKIPJACK': 0x0000660a,
    'CALG_HASH_REPLACE_OWF': 0x0000800b,
    'CALG_CYLINK_MEK': 0x0000660c,
    'CALG_MD4': 0x00008002,
    'CALG_AES_128': 0x0000660e,
    'CALG_AES_192': 0x0000660f,
    'CALG_AES_256': 0x00006610,
    'CALG_AES': 0x00006611,
    'CALG_AGREEDKEY_ANY': 0x0000aa03,
    'CALG_SHA1': 0x00008004,
    'CALG_MAC': 0x00008005,
    'CALG_MD2': 0x00008001,
    'CALG_TLS1_MASTER': 0x00004c06,
    'CALG_RSA_SIGN': 0x00002400,
    'CALG_SCHANNEL_ENC_KEY': 0x00004c07,
    'CALG_HMAC': 0x00008009,
    'CALG_TLS1PRF': 0x0000800a,
    'CALG_TEK': 0x0000660b,
    'CALG_SHA_256': 0x0000800c,
    'CALG_SHA_384': 0x0000800d,
    'CALG_SHA_512': 0x0000800e,
    'CALG_HUGHES_MD5': 0x0000a003,
    'CALG_RC4': 0x00006801,
    'CALG_ECDSA': 0x00002203,
    'CALG_RC2': 0x00006602,
    'CALG_SEAL': 0x00006802,
    'CALG_SSL3_MASTER': 0x00004c01,
    'CALG_SCHANNEL_MASTER_HASH': 0x00004c02,
    'CALG_MD5': 0x00008003,
    'CALG_SCHANNEL_MAC_KEY': 0x00004c03,
    'CALG_KEY_KEYX': 0x0000aa04,
    'CALG_ECMQV': 0x0000a001,
    'CALG_PCT1_MASTER': 0x00004c04,
    'CALG_RSA_KEYX': 0x0000a400,
    'CALG_OID_INFO_CNG_ONLY': 0xffffffff,
    'CALG_SSL2_MASTER': 0x00004c05,
    'CALG_OID_INFO_PARAMETERS': 0xfffffffe,
}


def AlgorithmID(subcon):
    r"""
    Converts an integer to an AlgorithmID value

    >>> str(AlgorithmID(Int16ul).parse("\x00\xa4"))
    'CALG_RSA_KEYX'
    >>> AlgorithmID(Int16ul).build("CALG_RC4")
    '\x01h'
    """
    return Enum(subcon, **ALGIDS)


"""This modules contains helper functions for the construct library."""

import os
import io
import re
import string
import uuid
import zlib

from mwcp.utils import custombase64, elffileutils, pefileutils


# Patch with version 2.8 changes.
from mwcp.utils.construct import version28 as construct
from .version28 import *


BYTE = Byte
WORD = Int16ul
DWORD = ULONG = Int32ul
QWORD = ULONGLONG = Int64ul


def chunk(seq, size):
    """
    Returns an iterator that yields full chunks seq into size chunks.

    >>> list(chunk('hello', 2))
    [('h', 'e'), ('l', 'l')]
    >>> list(chunk('hello!', 2))
    [('h', 'e'), ('l', 'l'), ('o', '!')]
    """
    return zip(*([iter(seq)] * size))


class Boolean(Adapter):
    r"""
    Adapter used to convert parsed value into a boolean.
    NOTE: While similar to construct.Flag, this adapter accepts any value other than 0 or '' as true.
          And will work with more than just construct.Byte.

    WARNING: Due to the lossy nature, this can't be used to build.

    e.g.
    >>> Boolean(Int32ul).parse(b'\x01\x02\x03\x04')
    True
    >>> Boolean(Int32ul).parse(b'\x00\x00\x00\x00')
    False
    >>> Boolean(CString()).parse(b'hello\x00')
    True
    >>> Boolean(CString()).parse(b'\x00')
    False
    """

    def _decode(self, obj, context, path):
        return bool(obj)


class ErrorMessage(Construct):
    r"""
    Raises an exception when triggered by parse or build. Can be used as a sentinel that blows a whistle when a conditional branch goes the wrong way, or to raise an error explicitly the declarative way.
    This modification allows the ability to supply a custom message.

    Example::

        >>> d = "x"/Int8sb >> IfThenElse(this.x > 0, Int8sb, ErrorMessage('Failed if statement'))
        >>> d.parse(b"\xff\x05")
        Traceback (most recent call last):
            ...
        ExplicitError: Failed if statement
    """
    __slots__ = ['message']

    def __init__(self, message="Error field was activated."):
        super(self.__class__, self).__init__()
        self.message = message

    def _parse(self, stream, context, path):
        message = self.message(context) if callable(self.message) else self.message
        raise ExplicitError(message)

    def _build(self, obj, stream, context, path):
        message = self.message(context) if callable(self.message) else self.message
        raise ExplicitError(message)


def String16(length):
    r"""
    Creates UTF-16 (little endian) encoded string.

    >>> String16(10).build(u'hello')
    'h\x00e\x00l\x00l\x00o\x00'
    >>> String16(10).parse(b'h\x00e\x00l\x00l\x00o\x00')
    u'hello'
    >>> String16(16).parse(b'h\x00e\x00l\x00l\x00o\x00\x00\x00\x00\x00\x00\x00')
    u'hello'
    """
    return String(length, encoding='utf-16-le')


def String32(length):
    r"""
    Creates UTF-32 (little endian) encoded string.

    >>> String32(20).build(u'hello')
    'h\x00\x00\x00e\x00\x00\x00l\x00\x00\x00l\x00\x00\x00o\x00\x00\x00'
    >>> String32(20).parse(b'h\x00\x00\x00e\x00\x00\x00l\x00\x00\x00l\x00\x00\x00o\x00\x00\x00')
    u'hello'
    """
    return String(length, encoding='utf-32-le')


class Printable(Validator):
    r"""
    Validator used to validate that a parsed String (or Bytes) is a printable (ascii) string.

    NOTE: A ValidationError is a type of ConstructError and will be cause if catching ConstructError.

    >>> Printable(String(5)).parse(b'hello')
    u'hello'
    >>> Printable(String(5)).parse(b'he\x11o!')
    Traceback (most recent call last):
        ...
    ValidationError: object failed validation: heo!
    >>> Printable(Bytes(3)).parse(b'\x01NO')
    Traceback (most recent call last):
        ...
    ValidationError: object failed validation: NO
    >>> Printable(Bytes(3)).parse(b'YES')
    'YES'
    """

    def _validate(self, obj, context, path):
        if isinstance(obj, bytes):
            return all(chr(byte) in string.printable for byte in obj)
        return isinstance(obj, stringtypes) and all(char in string.printable for char in obj)


# Continuously parses until it hits the first non-zero byte.
SkipNull = Const(b'\x00')[:]

# Continuously parses until it hits the first zero byte (consumed).
# Use this instead of CString() if you can't guarantee it won't fail to decode.
CBytes = NullTerminated(GreedyBytes)


class BytesTerminated(NullTerminated):
    r"""
    BytesTerminated is the same as NullTerminated except that it is targeted for binary data and not strings, and
    therefore the terminator can be an arbitrary length (as opposed to having length equal to the character width).
    See the NullTerminated documentation for the remainder of the functionality and options.

    >>> BytesTerminated(GreedyBytes, term=b'TERM').parse(b'helloTERM')
    'hello'
    """

    # The only method we need to override is _parse. Everything else from NullTerminated works as-is.
    def _parse(self, stream, context, path):
        term = self.term
        term_len = len(term)
        if term_len < 1:
            raise PaddingError("BytesTerminated term must be at least 1 byte")
        data = b''
        while True:
            pos = stream_tell(stream)
            try:
                b = stream_read(stream, term_len)
                stream_seek(stream, pos, 0)
            except StreamError:
                if self.require:
                    raise
                else:
                    stream_seek(stream, pos, 0)
                    data += stream_read_entire(stream)
                    break

            if b == term:
                if self.include:
                    data += b
                if self.consume:
                    stream_read(stream, term_len)
                break
            else:
                data += stream_read(stream, 1)
        if self.subcon is GreedyBytes:
            return data
        if type(self.subcon) is GreedyString:
            return data.decode(self.subcon.encoding)
        return self.subcon._parsereport(io.BytesIO(data), context, path)


# TODO: Make a LStripped?
class Stripped(Adapter):
    r"""
    An adapter that strips characters/bytes from the right of the parsed results.

    NOTE: While this may look similar to Padded() this is different because this
    doesn't take a length and instead strips out the nulls from within the already parsed subconstruct.

    :param subcon: The sub-construct to wrap.
    :param pad: The character/bytes to use for stripping. Defaults to null character.

    >>> Stripped(GreedyBytes).parse(b'hello\x00\x00\x00')
    'hello'
    >>> Stripped(Bytes(10)).parse(b'hello\x00\x00\x00\x00\x00')
    'hello'
    >>> Stripped(Bytes(14), pad=b'PAD').parse(b'helloPADPADPAD')
    'hello'
    >>> Stripped(Bytes(14), pad=b'PAD').build(b'hello')
    'helloPADPADPAD'
    >>> Stripped(CString(), pad=u'PAD').parse(b'helloPADPAD\x00')
    u'hello'
    >>> Stripped(String(14), pad=u'PAD').parse(b'helloPADPAD\x00\x00\x00')
    u'hello'

    # WARNING: If padding doesn't fit in the perscribed data it will not strip it!
    >>> Stripped(Bytes(13), pad=b'PAD').parse(b'helloPADPADPA')
    'helloPADPADPA'
    >>> Stripped(Bytes(13), pad=b'PAD').build(b'hello')
    Traceback (most recent call last):
        ...
    StreamError: bytes object of wrong length, expected 13, found 5

    # If the wrapped subconstruct's size can't be determine, if defaults to not providing a pad.
    >>> Stripped(CString(), pad=u'PAD').build(u'hello')
    'hello\x00'
    """

    def __init__(self, subcon, pad=None):
        super(Stripped, self).__init__(subcon)
        self.pad = pad

    def _decode(self, obj, context, path):
        pad = self.pad

        if pad is None:
            pad = u'\0' if isinstance(obj, unicodestringtype) else b'\x00'

        if not isinstance(pad, type(obj)):
            raise PaddingError("NullStripped pad must be of the same type: {} vs {}".format(type(pad), type(obj)))

        unit = len(pad)
        if unit < 1:
            raise PaddingError("NullStripped pad must be at least 1 byte")

        obj = obj
        if unit == 1:
            obj = obj.rstrip(pad)
        else:
            tailunit = len(obj) % unit
            end = len(obj)
            if tailunit and obj[-tailunit:] == pad[:tailunit]:
                end -= tailunit
            while end-unit >= 0 and obj[end-unit:end] == pad:
                end -= unit
            obj = obj[:end]

        return obj

    def _encode(self, obj, context, path):
        pad = self.pad

        if pad is None:
            pad = u'\0' if isinstance(self.subcon, StringEncoded) else b'\x00'

        try:
            size = self.subcon._sizeof(context, path)
        except SizeofError:
            return obj  # Don't pad if we can't figure out size.

        unit = len(pad)
        if unit == 1:
            obj = obj.ljust(size, pad)
        # Only pad if it fits in nicely.
        elif (size - len(obj)) % unit == 0:
            obj = (obj + (pad * (size - len(obj))))[:size]

        return obj


class HexString(Adapter):
    r"""
    Adapter used to convert an int into a hex string equivalent.

    e.g.
    >>> HexString(Int32ul).build('0x123')
    '#\x01\x00\x00'
    >>> HexString(Int32ul).parse(b'\x20\x01\x00\x00')
    '0x120'
    >>> HexString(Int16ub).parse(b'\x12\x34')
    '0x1234'
    >>> HexString(BytesInteger(20)).parse(b'\x01' * 20)
    '0x101010101010101010101010101010101010101'
    """

    def _encode(self, obj, context, path):
        return int(obj, 16)

    def _decode(self, obj, context, path):
        hex_string = hex(obj)
        if hex_string.endswith('L'):
            hex_string = hex_string[:-1]
        return hex_string


class Base64(Adapter):
    r"""
    Adapter used to Base64 encoded/decode a value.

    WARNING: This adapter must be used on a unicode string value.

    :param subcon: the construct to wrap
    :param custom_alpha: optional custom alphabet to use

    e.g.
    >>> Base64(GreedyString()).build(b'hello')
    'aGVsbG8='
    >>> Base64(GreedyString()).parse(b'aGVsbG8=')
    'hello'
    >>> Base64(GreedyBytes).build(b'\x01\x02\x03\x04')
    'AQIDBA=='
    >>> Base64(GreedyBytes).parse(b'AQIDBA==')
    '\x01\x02\x03\x04'

    NOTE: String size is based on the encoded version.
    >>> Base64(String(16)).build(u'hello world')
    'aGVsbG8gd29ybGQ='
    >>> Base64(String(16)).parse(b'aGVsbG8gd29ybGQ=')
    'hello world'

    Supplying a custom alphabet is also supported.
    >>> spec = Base64(String(16), custom_alpha='EFGHQRSTUVWefghijklmnopIJKLMNOPABCDqrstuvwxyXYZabcdz0123456789+/=')
    >>> spec.build(u'hello world')
    'LSoXMS8BO29dMSj='
    >>> spec.parse(b'LSoXMS8BO29dMSj=')
    'hello world'
    """
    __slots__ = ['subcon', 'custom_alpha']

    def __init__(self, subcon, custom_alpha=None):
        super(Base64, self).__init__(subcon)
        self.custom_alpha = custom_alpha

    def _encode(self, obj, context, path):
        obj = custombase64.b64encode(obj, alphabet=self.custom_alpha)
        # Convert to unicode if wrapped subcon expects it.
        if isinstance(self.subcon, StringEncoded):
            obj = obj.decode('utf-8')
        return obj

    def _decode(self, obj, context, path):
        if isinstance(obj, str):
            obj = obj.encode('utf-8')
        return custombase64.b64decode(obj, alphabet=self.custom_alpha)


class ZLIB(Adapter):
    r"""
    Adapter used to zlib compress/decompress a data buffer

    :param subcon: The construct to wrap
    :param int level: The zlib compression level
    :param int wbits: The zlib decompression window size
    :param int bufsize: The initial output buffer size

    >>> ZLIB(Bytes(12)).build(b'data')
    'x\x9cKI,I\x04\x00\x04\x00\x01\x9b'
    >>> ZLIB(GreedyBytes, level=0).build(b'data')
    'x\x01\x01\x04\x00\xfb\xffdata\x04\x00\x01\x9b'
    >>> ZLIB(GreedyBytes).parse(b'x^KI,I\x04\x00\x04\x00\x01\x9b')
    'data'
    """
    __slots__ = ["subcon", "wbits", "bufsize", "level"]

    def __init__(self, subcon, wbits=None, bufsize=None, level=None):
        super(ZLIB, self).__init__(subcon)
        self.wbits = wbits
        self.bufsize = bufsize
        self.level = level

    def _encode(self, obj, context, path):
        level = self.level(context) if callable(self.level) else self.level
        if level is not None:
            return zlib.compress(obj, level)
        return zlib.compress(obj)

    def _decode(self, obj, context, path):
        """
        ZLIB decompress a buffer, cannot use bufsize if wbits is not set

        :param obj:
        :param context:

        :return:
        """
        wbits = self.wbits(context) if callable(self.wbits) else self.wbits
        bufsize = self.bufsize(context) if callable(self.bufsize) else self.bufsize
        if wbits is not None and bufsize is not None:
            return zlib.decompress(obj, wbits, bufsize)
        elif wbits is not None:
            return zlib.decompress(obj, wbits)
        return zlib.decompress(obj)


class UUIDAdapter(Adapter):
    r"""
    Adapter used to convert parsed bytes to a string representing the UUID.
    Adapter can decode 16 bytes straight or in little-endian order if you set le=True.

    e.g.
    >>> UUIDAdapter(Bytes(16)).build('{12345678-1234-5678-1234-567812345678}')
    'xV4\x124\x12xV\x124Vx\x124Vx'
    >>> UUIDAdapter(Bytes(16), le=False).build('{12345678-1234-5678-1234-567812345678}')
    '\x124Vx\x124Vx\x124Vx\x124Vx'
    >>> UUIDAdapter(Bytes(16)).parse(b'xV4\x124\x12xV\x124Vx\x124Vx')
    '{12345678-1234-5678-1234-567812345678}'
    """
    __slots__ = ['subcon', 'le']

    def __init__(self, subcon, le=True):
        super(UUIDAdapter, self).__init__(subcon)
        self.le = le

    def _encode(self, obj, context, path):
        obj = uuid.UUID(obj)
        if self.le:
            return obj.bytes_le
        else:
            return obj.bytes

    def _decode(self, obj, context, path):
        if self.le:
            _uuid = uuid.UUID(bytes_le=obj)
        else:
            _uuid = uuid.UUID(bytes=obj)
        return '{' + str(_uuid) + '}'


def UUID(le=True):
    r"""A convenience function for using the UUIDAdapter with 16 bytes.

    :param le: Whether to use "bytes_le" or "bytes" when constructing the UUID.

    e.g.
    >>> UUID().build('{12345678-1234-5678-1234-567812345678}')
    'xV4\x124\x12xV\x124Vx\x124Vx'
    >>> UUID(le=False).build('{12345678-1234-5678-1234-567812345678}')
    '\x124Vx\x124Vx\x124Vx\x124Vx'
    >>> UUID().parse(b'xV4\x124\x12xV\x124Vx\x124Vx')
    '{12345678-1234-5678-1234-567812345678}'
    >>> UUID(le=False).parse(b'\x124Vx\x124Vx\x124Vx\x124Vx')
    '{12345678-1234-5678-1234-567812345678}'
    """
    return UUIDAdapter(Bytes(16), le=le)


def ELFPointer(mem_off, subcon, elf=None):
    r"""
    Pointer for ELF files. This works for both memory sizes.

    NOTE: This only works for x86 instructions. For other architectures,
    please see the "ELFPointer" within their respective submodules. (e.g. construct.ARM.ELFPointer)

    spec.parse(file_data, pe=elf_object)

    :param mem_off: an int or a function that represents the memory offset for the equivalent physical offset.
    :param subcon: the subcon to use at the offset
    :param elf: Optional elftools.ELFFile file object.
        (if not supplied here, this must be supplied during parse()/build()
    """
    def _obtain_physical_offset(ctx):
        _elf = elf or ctx._params.elf
        _mem_off = mem_off(ctx) if callable(mem_off) else mem_off
        phy_off = elffileutils.obtain_physical_offset(_mem_off, elf=_elf)
        if phy_off is None:
            raise ConstructError('Unable to decode virtual address')
        return phy_off

    return Pointer(_obtain_physical_offset, subcon)


class PEPhysicalAddress(Adapter):
    r"""
    Adapter used to convert an int representing a PE memory address into a physical address.

    The PE object can either be passed into the specific construct, or as a keyword arument in
    the parse()/build() functions.
    If passed in through parse()/build(), the same PE object will be used for all instances.

    This Adapter is useful when used along-side the Pointer construct:
    spec = Struct(
        'offset' / PEPhysicalAddress(Int32ul),
        'data' / Pointer(this.offset, Bytes(100))
    )

    e.g.
    >> with open(r'C:\32bit_exe', 'rb') as fo:
    ...     file_data = fo.read()
    >> pe = pefileutils.obtain_pe(file_data)
    >> PEPhysicalAddress(Int32ul, pe=pe).build(100)
    'd\x00@\x00'
    >> PEPhysicalAddress(Int32ul, pe=pe).parse(b'd\x00@\x00')
    100
    >> PEPhysicalAddress(Int32ul).build(100, pe=pe)
    'd\x00@\x00'
    >> PEPhysicalAddress(Int32ul).parse(b'd\x00@\x00', pe=pe)
    100
    """
    def __init__(self, subcon, pe=None):
        """
        :param pe: Optional PE file object. (if not supplied here, this must be supplied during parse()/build()
        :param subcon: subcon to parse memory offset.
        """
        super(PEPhysicalAddress, self).__init__(subcon)
        self._pe = pe

    def _encode(self, obj, context, path):
        pe = self._pe or context._params.pe
        address = pefileutils.obtain_memory_offset(obj, pe=pe)
        if address is None:
            raise ConstructError('Unable to encode physical address.')
        return address

    def _decode(self, obj, context, path):
        pe = self._pe or context._params.pe
        address = pefileutils.obtain_physical_offset(obj, pe=pe)
        if address is None:
            raise ConstructError('Unable to decode virtual address.')
        return address


def PEPointer(mem_off, subcon, pe=None):
    r"""
    This is an alternative to PEPhysicalAddress when you are using the address along with Pointer

    Simplifies:
    spec = Struct(
        'offset' / PEPhysicalAddress(Int32ul),
        'data' / Pointer(this.offset, Bytes(100))
    )
    to:
    spec = Struct(
        'offset' / Int32ul,
        'data' / PEPointer(this.offset, Bytes(100))
    )

    spec.parse(file_data, pe=pe_object)

    :param mem_off: an int or a function that represents the memory offset for the equivalent physical offset.
    :param subcon: the subcon to use at the offset
    :param pe: Optional PE file object. (if not supplied here, this must be supplied during parse()/build()
    """
    def _obtain_physical_offset(ctx):
        _pe = pe or ctx._params.pe
        _mem_off = mem_off(ctx) if callable(mem_off) else mem_off
        phy_off = pefileutils.obtain_physical_offset(_mem_off, pe=_pe)
        if phy_off is None:
            raise ConstructError('Unable to decode virtual address')
        return phy_off

    return Pointer(_obtain_physical_offset, subcon)


def PEPointer64(mem_off, inst_end, subcon, pe=None):
    r"""
    This is the 64-bit version of PEPointer.
    This subconstruct takes an extra argument which specifies
    the location of the end of the instruction for which the memory_offset was used.
    (A parameter necessary for 64-bit)

    Example:
    spec = Struct(
        'offset' / Int32ul,
        Padding(2),
        'inst_end' / Tell,
        'data' / PEPointer64(this.offset, this.inst_end, Byte(100))
    )

    spec = Struct(
        'instruction' / Regex(
            '\x01\x03(?P<data_ptr>.{4})\x04\x05(?P<end>)\x06\x07', data_ptr=DWORD, end=Tell),
        'data' / PEPointer64(this.instruction.data_ptr, this.instruction.end, Bytes(100))
    )

    spec.parse(file_data, pe=pe_object)

    :param mem_off: an int or a function that represents the memory offset for the equivelent physical offset.
    :param inst_end: an int or a function that represents the location of the end of the instruction to be relative to.
    :param subcon: the subcon to use at the offset
    :param pe: Optional PE file object. (if not supplied here, this must be supplied during parse()/build()
    """
    def _obtain_physical_offset(ctx):
        _pe = pe or ctx._params.pe
        _mem_off = mem_off(ctx) if callable(mem_off) else mem_off
        _inst_end = inst_end(ctx) if callable(inst_end) else inst_end
        phy_off = pefileutils.obtain_physical_offset_x64(_mem_off, _inst_end, pe=_pe)
        if phy_off is None:
            raise ConstructError('Unable to decode virtual address')
        return phy_off

    return Pointer(_obtain_physical_offset, subcon)


class Delimited(Construct):
    r"""
    A construct used to parse delimited data.

    NOTE: The parsed constructs will be buffered

    >>> spec = Delimited(b'|',
    ...     'first' / CString(),
    ...     'second' / Int32ul,
    ...     # When using a Greedy construct, either all data till EOF or the next delimiter will be consumed.
    ...     'third' / GreedyBytes,
    ...     'fourth' / Byte
    ... )
    >>> spec.parse(b'Hello\x00\x00|\x01\x00\x00\x00|world!!\x01\x02|\xff')
    Container(first=u'Hello', second=1, third=b'world!!\x01\x02', fourth=255)
    >>> spec.build(dict(first=u'Hello', second=1, third=b'world!!\x01\x02', fourth=255))
    'Hello\x00|\x01\x00\x00\x00|world!!\x01\x02|\xff'

    If you don't care about a particular element, you can leave it nameless just like in Structs.
    # NOTE: You can't build unless you have supplied every attribute.
    >>> spec = Delimited(b'|',
    ...     'first' / CString(),
    ...     'second' / Int32ul,
    ...     Pass,
    ...     'fourth' / Byte
    ... )
    >>> spec.parse(b'Hello\x00\x00|\x01\x00\x00\x00|world!!\x01\x02|\xff')
    Container(first=u'Hello', second=1, fourth=255)

    It may also be useful to use Pass or Optional for fields that may not exist.
    >>> spec = Delimited(b'|',
    ...     'first' / CString(),
    ...     'second' / Pass,
    ...     'third' / Optional(Int32ul)
    ... )
    >>> spec.parse(b'Hello\x00\x00|dont care|\x01\x00\x00\x00')
    Container(first=u'Hello', second=None, third=1)
    >>> spec.parse(b'Hello\x00\x00||')
    Container(first=u'Hello', second=None, third=None)

    delimiters may have a length > 1
    >>> spec = Delimited(b'YOYO',
    ...     'first' / CString(),
    ...     'second' / Int32ul,
    ...     # When using a Greedy construct, either all data till EOF or the next delimiter will be consumed.
    ...     'third' / GreedyBytes,
    ...     'fourth' / Byte
    ... )
    >>> spec.parse(b'Hello\x00\x00YOYO\x01\x00\x00\x00YOYOworld!!YO!!\x01\x02YOYO\xff')
    Container(first=u'Hello', second=1, third=b'world!!YO!!\x01\x02', fourth=255)
    >>> spec.build(dict(first=u'Hello', second=1, third=b'world!!YO!!\x01\x02', fourth=255))
    'Hello\x00YOYO\x01\x00\x00\x00YOYOworld!!YO!!\x01\x02YOYO\xff'

    # TODO: Add support for using a single construct for parsing an unknown number of times
    # (or within a min, max, or exact)
    # (Perhaps call it "Split" to avoid overloading too much functionality.)
    # e.g.
    # >>> spec = Delimited(b'|', GreedyString())
    # >>> spec.parse(b'hello|world')
    # ['hello', 'world']
    # >>> spec.parse(b'hello|world|hi|bob')
    # ['hello', 'world', 'hi', 'bob']
    # >>> spec.parse(b'hello')
    # ['hello']
    """

    __slots__ = ['delimiter', 'subcons']

    def __init__(self, delimiter, *subcons):
        """
        :param delimiter: single charactor or a function that takes context and returns the delimiter
        :param subcons: constructs to use to parse each element.
                    NOTE: The number of constructs will be the number of elements delimited.
                    (ie. len(subcons) == number of delimiters + 1)

        :raises ValueError: If no subcons are defined.
        """
        super(Delimited, self).__init__()
        self.delimiter = delimiter
        self.subcons = subcons
        if len(subcons) < 2:
            raise ValueError('At least two subconstruct must be defined.')

    def _find_delimiter(self, stream, delimiter):
        """
        Finds given delimiter in stream.

        :returns: Stream offset for delimiter.
        :raises ConstructError: If delimiter isn't found.
        """
        fallback = stream.tell()
        try:
            for byte in iter(lambda: stream.read(1), b''):
                if delimiter[0] == ord(byte):
                    delimiter_offset = stream.seek(-1, os.SEEK_CUR)
                    if stream.read(len(delimiter)) == delimiter:
                        return delimiter_offset
                    else:
                        stream.seek(delimiter_offset + 1)
            raise ConstructError('Unable to find delimiter: {}'.format(delimiter))
        finally:
            stream.seek(fallback)

    def _parse_subcon(self, subcon, stream, obj, context, path):
        """Parses and fills obj and context."""
        subobj = subcon._parsereport(stream, context, path)
        if subcon.flagembedded:
            if subobj is not None:
                obj.update(subobj.items())
                context.update(subobj.items())
        else:
            if subcon.name is not None:
                obj[subcon.name] = subobj
                context[subcon.name] = subobj

    def _parse(self, stream, context, path):
        delimiter = self.delimiter(context) if callable(self.delimiter) else self.delimiter
        if not isinstance(delimiter, bytestringtype) or not delimiter:
            raise ValueError('Invalid delimiter.')

        obj = Container()
        context = Container(_=context)

        # Parse all but the last element.
        for sc in self.subcons[:-1]:
            # Don't count probes as an element.
            if isinstance(sc, Probe):
                sc._parsereport(stream, context, path)
                continue

            delimiter_offset = self._find_delimiter(stream, delimiter)

            # Temporaily fake the read() so that we can force EOF before delimiter.
            orig_read = stream.read
            def new_read(size=None):
                max_size = delimiter_offset - stream.tell()
                if size is None:
                    size = max_size
                else:
                    size = min(max_size, size)
                return orig_read(size)
            try:
                stream.read = new_read
                self._parse_subcon(sc, stream, obj, context, path)
            finally:
                stream.read = orig_read

            # Align to after delimiter
            stream.seek(delimiter_offset + len(delimiter))

        # Parse the last element.
        self._parse_subcon(self.subcons[-1], stream, obj, context, path)

        return obj

    def _build(self, obj, stream, context, path):
        delimiter = self.delimiter(context) if callable(self.delimiter) else self.delimiter
        if not isinstance(delimiter, bytestringtype) or not delimiter:
            raise ValueError('Invalid delimiter.')

        context = Container(_=context)
        context.update(obj)
        for i, sc in enumerate(self.subcons):
            if sc.flagembedded:
                subobj = obj
            elif sc.flagbuildnone:
                subobj = obj.get(sc.name, None)
            else:
                subobj = obj[sc.name]
            buildret = sc._build(subobj, stream, context, path)
            if buildret is not None:
                if sc.flagembedded:
                    context.update(buildret)
                if sc.name is not None:
                    context[sc.name] = buildret
            # Add delimiter if not last element and not Probe.
            if i < len(self.subcons) - 1 and not isinstance(sc, Probe):
                stream.write(delimiter)
        return context


class Regex(Construct):
    r"""
    A construct designed look for the first match for the given regex, then parse the data collected in the groups.
    Returns the matched capture groups in attributes based on their respective names.
    If a subconstruct is defined for a group, it will run that construct on that particular piece of data.

    NOTE: The subconstruct will run on the data as if is the only data that exists. Therefore, using Seek and Tell
    will be purely relative to that piece of data only. This was done to ensure you are only parsing what has been
    captured. (If you need to use Seek or Tell, you will have to instead make a capture group that collects no data.)


    NOTE: If you supply a string as the regular expression, the re.DOTALL flag will be automatically specified.
    If you need to use different flags, you must past a compiled regex.

    The seek position is left at the end of the successful match (match.end()).

    >>> regex = re.compile('\x01\x02(?P<size>.{4})\x03\x04(?P<path>[A-Za-z].*\x00)', re.DOTALL)
    >>> data = 'GARBAGE!\x01\x02\x0A\x00\x00\x00\x03\x04C:\Windows\x00MORE GARBAGE!'
    >>> r = Regex(regex, size=Int32ul, path=CString()).parse(data)
    >>> r == Container(path=u'C:\\Windows', size=10)
    True
    >>> r = Regex(regex).parse(data)
    >>> r == Container(path=b'C:\\Windows\x00', size=b'\n\x00\x00\x00')
    True
    >>> r = Struct(
    ...     're' / Regex(regex, size=Int32ul, path=CString()),
    ...     'after_re' / Tell,
    ...     'garbage' / GreedyBytes
    ... ).parse(data)
    >>> r == Container(re=Container(path=u'C:\\Windows', size=10), after_re=27L, garbage=b'MORE GARBAGE!')
    True

    # TODO: Unfortunately Embedded() no longer works with the update to 2.9
    # >>> Struct(
    # ...     Embedded(Regex(regex, size=Int32ul, path=CString())),
    # ...     'after_re' / Tell,
    # ...     'garbage' / GreedyBytes
    # ... ).parse(data)
    # Container(path=u'C:\\Windows', size=10, after_re=27L, garbage=b'MORE GARBAGE!')

    You can use Regex as a trigger to find a particular piece of data before you start parsing.
    >>> Struct(
    ...     Regex('TRIGGER'),
    ...     'greeting' / CString()
    ... ).parse('\x01\x02\x04GARBAGE\x05TRIGGERhello world\x00')
    Container(greeting=u'hello world')

    If no data is captured, the associated subcon will received a stream with the position set at the location
    of that captured group. Thus, allowing you to use it as an anchor point.
    >>> r = Regex('hello (?P<anchor>)world(?P<extra_data>.*)', anchor=Tell).parse('hello world!!!!')
    >>> r == Container(extra_data=b'!!!!', anchor=6L)
    True

    If no named capture groups are used, you can instead parse the entire matched string by supplying
    a subconstruct as a positional argument. (If no subcon is provided, the raw bytes are returned instead.
    >>> Regex('hello world\x00', CString()).parse('GARBAGE\x01\x03hello world\x00\x04')
    u'hello world'
    >>> Regex('hello world\x00').parse('GARBAGE\x01\x03hello world\x00\x04')
    'hello world\x00'

    You can also set the regular expression to match in-place (instead of searching the data)
    by setting the keyword argument _match to True.
    >>> Regex('hello', _match=True).parse(b'hello world!')
    'hello'
    >>> Regex('hello').parse(b'bogus hello world')
    'hello'
    >>> Regex('hello', _match=True).parse(b'bogus hello world')
    Traceback (most recent call last):
        ...
    ConstructError: [(parsing)] regex did not match
    """

    __slots__ = ['regex', 'subcon', 'group_subcons', 'match']

    def __init__(self, regex, *subcon, **group_subcons):
        """
        Initializes regex construct.

        :param regex: A regex to use (can be a string or compiled).
        :param subcon:
            A subcon to use on the entire matching string when there are no named capture groups.
            (NOTE: This is only used if there are no capture groups.
            If you want to use capture groups AND this then have a capture group encapsulating the entire regex.)
        :param group_subcons:
            Keyword argument dictionary that contains the constructs to use for the corresponding capture group.
            If a subcon is not supplied for a capture group, it will default to returning bytes
            (equivalent to setting construct.Bytes() for that group.)

        :raises ValueError: If arguments are invalid.
        """
        super(Regex, self).__init__()
        if isinstance(regex, str):
            regex = regex.encode()  # force byte strings
        if isinstance(regex, bytestringtype):
            regex = re.compile(regex, re.DOTALL)
        self.regex = regex
        # TODO: This feature seems backwards, perhaps make a _search keyword instead and default to match functionality.
        # Alternatively, we could have RegexSearch and RegexMatch constructs instead.
        self.match = group_subcons.pop('_match', False)
        self.group_subcons = group_subcons
        if subcon and len(subcon) > 1:
            raise ValueError('Only one subcon can be supplied for the entire match.')
        if subcon and group_subcons:
            raise ValueError('subcon and group_subcons arguments cannot be used at the same time.')
        self.subcon = subcon[0] if subcon else None

    def _parse(self, stream, context, path):
        start = stream.tell()
        # NOTE: we are going to have to read the entire stream due to regex requirements.
        # However, that's okay in this case since we are parsing ByteIO anyway.
        if self.match:
            match = self.regex.match(stream.read())
        else:
            match = self.regex.search(stream.read())
        if not match:
            raise ConstructError('[{}] regex did not match'.format(path))

        try:
            group_dict = match.groupdict()

            # If there are no named groups. Return parsed full match instead.
            if not group_dict:
                if self.subcon:
                    sub_stream = io.BytesIO(match.group())
                    return self.subcon._parsereport(sub_stream, context, path)
                else:
                    return match.group()

            # Otherwise, we are going to parse each named capture group.
            obj = Container()
            obj._io = stream

            context = Container(_=context, _params=context._params, _root=None, _parsing=context._parsing,
                                _building=context._building, _sizing=context._sizing, _subcons=self.group_subcons,
                                _io=stream, _index=context.get("_index", None))
            context._root = context._.get("_root", context)

            # Default to displaying matched data as pure bytes.
            obj.update(group_dict)
            context.update(group_dict)

            # Parse groups using supplied constructs.
            for name, subcon in self.group_subcons.items():
                try:
                    data = match.group(name)
                except IndexError:
                    continue

                # If data is None, then we are most likely dealing with an optional capture group.
                if data is None:
                    obj[name] = None
                    context[name] = None
                    continue

                # If we have an empty capture group, the user would like to use it as an anchor.
                if not data:
                    stream.seek(start + match.start(name))
                    sub_stream = stream
                else:
                    sub_stream = io.BytesIO(data)

                try:
                    subobj = subcon._parsereport(sub_stream, context, path)
                except ConstructError as e:
                    # Raise a more useful error message.
                    # TODO: Remove when path is provided in exception messages.
                    raise ConstructError('Failed to parse {} capture group with error: {}'.format(name, e))
                obj[name] = subobj
                context[name] = subobj
            return obj

        finally:
            # Reset position to right after the matched regex.
            stream.seek(start + match.end())

    def _build(self, obj, stream, context, path):
        raise ConstructError('Building for Regex is not supported.')

    def _sizeof(self, context, path):
        raise SizeofError('sizeof() for Regex is not supported.')

    def _emitparse(self, code):
        raise NotImplementedError

    def _emitseq(self, ksy, bitwise):
        raise NotImplementedError


class IterError(ConstructError):
    pass


# TODO: Should this be renamed to Map?
class Iter(Construct):
    r"""
    Class that allows iterating over an object and acting on each item.

    e.g.
    >>> spec = Struct(
    ...     'types' / Byte[3],
    ...     'entries' / Iter(this.types, {
    ...        1: construct.Int32ul,
    ...        2: construct.Int16ul,
    ...     },
    ...     default=construct.Pass
    ...     )
    ... )
    >>> spec.parse('\x01\x02\x09\x03\x03\x03\x03\x06\x06')
    Container(types=ListContainer([1, 2, 9]), entries=ListContainer([50529027, 1542, None]))
    >>> C = _
    >>> spec.build(C)
    '\x01\x02\t\x03\x03\x03\x03\x06\x06'
    >>> spec.sizeof(**C)
    9

    >>> spec = Struct(
    ...     'sizes' / Int16ul[4],
    ...     'entries' / Iter(this.sizes, Bytes)  # equivalent to Iter(this.sizes, lambda size: Bytes(size))
    ... )
    >>> spec.parse('\x01\x00\x03\x00\x00\x00\x05\x00abbbddddd')
    Container(sizes=ListContainer([1, 3, 0, 5]), entries=ListContainer(['a', 'bbb', '', 'ddddd']))
    >>> C = _
    >>> spec.build(C)
    '\x01\x00\x03\x00\x00\x00\x05\x00abbbddddd'
    >>> Iter(this.sizes, Bytes).sizeof(sizes=[1,2,3,0])
    6
    >>> spec.sizeof(**C)
    17

    :param iterable: iterable items to act upon
    :param cases: A dictionary of cases or a function that takes takes a key and returns a construct spec.
    :param default: The default case (only if cases is a dict)
    """
    __slots__ = ['iterable', 'cases', 'default']
    
    def __init__(self, iterable, cases, default=None):
        super(Iter, self).__init__()
        self.iterable = iterable
        self.cases = cases
        self.default = default or Pass
        if not callable(cases):
            self.flagbuildnone = all(sc.flagbuildnone for sc in cases.values())
            self.flagembedded = all(sc.flagembedded for sc in cases.values())
        
    def _parse(self, stream, context, path):
        iterator = iter(self.iterable(context)) if callable(self.iterable) else iter(self.iterable)
        if callable(self.cases):
            return ListContainer([self.cases(key)._parsereport(stream, context, path) for key in iterator])
        else:
            return ListContainer([self.cases.get(key, self.default)._parsereport(stream, context, path) for key in iterator])
        
    def _build(self, obj, stream, context, path):
        iterator = iter(self.iterable(context)) if callable(self.iterable) else iter(self.iterable)
        for sub_obj, key in zip(obj, iterator):
            if callable(self.cases):
                self.cases(key)._build(sub_obj, stream, context, path)
            else:
                self.cases.get(key, self.default)._build(sub_obj, stream, context, path)
            
    def _sizeof(self, context, path):
        iterator = iter(evaluate(self.iterable, context))
        size = 0
        for key in iterator:
            try:
                if callable(self.cases):
                    size += self.cases(key)._sizeof(context, path)
                else:
                    size += self.cases.get(key, self.default)._sizeof(context, path)
            except (KeyError, AttributeError):
                raise SizeofError("cannot calculate size, {!r} key not found in context".format(key))
        return size


def find_constructs(struct, data):
    r"""
    Generator that yields the results of successful parsings of the given
    construct.
    Note: Construct must attempt to read something. Ie, don't have a Peek
    as your first subconstruct.

    Also, it's best if you have some type of validation (Const, OneOf, NoneOf, Check, etc) within your struct.
    Otherwise, it makes more sense to use a GreedyRange (the '[:]' notation) instead of this function.

    e.g.
    >>> struct = Struct(
    ...     Const(b'MZ'),
    ...     'int' / Int16ul,
    ...     'string' / CString())
    >>> list(find_constructs(struct, b'\x01\x02\x03MZ\x0A\x00hello\x00\x03\x04MZ\x0B\x00world\x00\x00'))
    [(3L, Container(int=10, string=u'hello')), (15L, Container(int=11, string=u'world'))]
    >>> list(find_constructs(struct, b'nope'))
    []

    :param struct: construct to apply (instance of construct.Construct)
    :param data: byte string of data to search.

    :yield: tuple containing (offset with data, result Container class)
    """
    data = io.BytesIO(data)

    while True:
        offset = data.tell()
        try:
            data_element = struct.parse_stream(data)
        except (construct.ConstructError, OverflowError) as e:
            data.seek(offset + 1)
        else:
            yield offset, data_element

        # Test if we hit end of data.
        if data.read(1):
            data.seek(-1, os.SEEK_CUR)
        else:
            break


class Backwards(Subconstruct):
    r"""
    Subconstruct used to parse a given subconstruct backwards in the stream.
    This ia a macro for seeking backwards before parsing the construct.
    (This will not work for subcons that don't have a valid sizeof.
    Except for GreedyBytes and GreedyString)

    The stream will be left off at the start of the parsed result by design.
    Therefore, doing something like Int32ul >> Backwards(Int32ul) >> Int32ul will parse
    the same data 3 times.

    e.g.
    >>> (Bytes(14) >> Backwards(Int32ul) >> Tell).parse(b'junk stuff\x01\x02\x00\x00')
    ListContainer(['junk stuff\x01\x02\x00\x00', 513, 10L])
    >>> spec = Struct(Seek(0, os.SEEK_END), 'name' / Backwards(String(9)), 'number' / Backwards(Int32ul))
    >>> spec.parse(b'A BUNCH OF JUNK DATA\x01\x00\x00\x00joe shmoe')
    Container(name=u'joe shmoe', number=1)

    WARNING: This will break if the subcon doesn't have a valid sizeof.
    >>> spec = Struct(Seek(0, os.SEEK_END), 'name' / Backwards(CString()), 'number' / Backwards(Int32ul))
    >>> spec.parse(b'A BUNCH OF JUNK DATA\x01\x00\x00\x00joe shmoe\x00')
    Traceback (most recent call last):
      ...
    SizeofError

    However, GreedyBytes and GreedyString are allowed.
    >>> spec = Struct(Seek(0, os.SEEK_END), 'name' / Backwards(String(9)), 'rest' / Backwards(GreedyBytes))
    >>> spec.parse(b'A BUNCH OF JUNK DATA\x01\x00\x00\x00joe shmoe')
    Container(name=u'joe shmoe', rest=b'A BUNCH OF JUNK DATA\x01\x00\x00\x00')
    >>> spec = Struct(Seek(0, os.SEEK_END), 'name' / Backwards(String(9)), 'rest' / Backwards(GreedyString(encoding='utf-16-le')))
    >>> spec.parse(b'h\x00e\x00l\x00l\x00o\x00joe shmoe')
    Container(name=u'joe shmoe', rest=u'hello')

    WARNING: This will also break if you read more data that is behind the current position.
    >>> (Seek(0, os.SEEK_END) >> Backwards(String(10))).parse('yo')
    Traceback (most recent call last):
      ...
    FormatFieldError: could not read enough bytes, expected 10, found 2
    """
    __slots__ = ['greedy']

    def __init__(self, subcon):
        super(Backwards, self).__init__(subcon)
        # GreedyBytes and GreedyString are allowed special cases.
        self.greedy = self.subcon is GreedyBytes or (
                isinstance(self.subcon, construct.StringEncoded) and self.subcon.subcon is GreedyBytes)

    def _parse(self, stream, context, path):
        # Seek back to start of subcon.
        orig_pos = stream.tell()
        if self.greedy:
            start_pos = stream.seek(0)
            size = orig_pos - start_pos
            try:
                sub_stream = io.BytesIO(stream_read(stream, size))
                return self.subcon._parsereport(sub_stream, context, path)
            finally:
                stream.seek(start_pos)
        else:
            size = self.subcon._sizeof(context, path)
            start_pos = stream.seek(size * -1, os.SEEK_CUR)
            # Determine if we fell off the front.
            if orig_pos - start_pos < size:
                raise FormatFieldError("could not read enough bytes, expected %d, found %d" % (size, orig_pos - start_pos))
            try:
                return self.subcon._parsereport(stream, context, path)
            finally:
                stream.seek(start_pos)

    def _build(self, obj, stream, context, path):
        # TODO: Add support for building.
        raise NotImplementedError('Building is not supported.')


# Monkey patch RawCopy so that it can handle when we read the stream backwards.
def _parse(self, stream, context, path):
    offset1 = stream.tell()
    obj = self.subcon._parsereport(stream, context, path)
    offset2 = stream.tell()
    # Swap if subcon read backwards.
    if offset1 > offset2:
        offset1, offset2 = offset2, offset1
    fallback = stream.tell()
    stream_seek(stream, offset1)
    data = stream_read(stream, offset2-offset1)
    stream.seek(fallback)
    return Container(data=data, value=obj, offset1=offset1, offset2=offset2, length=(offset2-offset1))


construct.RawCopy._parse = _parse


def FocusLast(*subcons, **kw):
    r"""
    A helper for performing the common technique of using FocusedSeq to
    parse a bunch of subconstructs and then grab the last element.

    >>> FocusLast(Byte, Byte, String(2)).parse(b'\x01\x02hi')
    u'hi'

    >>> spec = FocusLast(
    ...     'a' / Byte,
    ...     'b' / Byte,
    ...     String(this.a + this.b),
    ... )
    >>> spec.parse(b'\x01\x02hi!')
    u'hi!'
    >>> spec.build(u'hi!', a=1, b=2)
    '\x01\x02hi!'


    e.g.:
        # Simplifies this:
        construct.FocusedSeq(
            'value',
            're' / construct.Regex(.., offset=construct.Int32ul, size=construct.Byte),
            'value' / construct.PEPointer(this.re.offset, construct.Bytes(this.re.size)
        )
        # To this:
        construct.FocusLast(
            're' / construct.Regex(.., offset=construct.Int32ul, size=construct.Byte),
            construct.PEPointer(this.re.offset, construct.Bytes(this.re.size)
        )
    """
    return FocusedSeq(len(subcons) - 1, *subcons, **kw)



"""
Network constructs
"""

from .version28 import *


class _MACAddressAdapter(Adapter):
    r"""
    Adapter used to format a MAC address from a list of 6 bytes

    e.g.
    >>> _MACAddressAdapter(Byte[6]).parse(b'\x00\x0c\x29\xd3\x91\xbc')
    '00-0c-29-d3-91-bc'
    """
    def _encode(self, obj, context, path):
        return list(map(chr, obj.split("-")))

    def _decode(self, obj, context, path):
        return '{:02x}-{:02x}-{:02x}-{:02x}-{:02x}-{:02x}'.format(*obj)


# A MacAddress parsed from single bytes.
MacAddress = _MACAddressAdapter(Byte[6])


class _IP4AddressAdapter(Adapter):
    r"""
    Adapter used to format a IP address from a list of four ints.

    e.g.
    >>> _IP4AddressAdapter(Byte[4]).parse(b'\x01\x02\x03\x04')
    '1.2.3.4'
    >>> _IP4AddressAdapter(Int16ul[4]).parse(b'\x01\x00\x02\x00\x03\x00\x04\x00')
    '1.2.3.4'
    """

    def _encode(self, obj, context, path):
        return list(map(int, obj.split('.')))

    def _decode(self, obj, context, path):
        return '{0}.{1}.{2}.{3}'.format(*obj)


# An IP4Address parsed from single bytes.
IP4Address = _IP4AddressAdapter(Byte[4])



"""This module is used to convert constructs to an HTML document.

To use, run the html_hex with a construct and data:
    print html_hex(CONSTRUCT, data)
"""
from __future__ import print_function

import codecs
import os
import construct
import itertools
import jinja2
import sys


PY3 = sys.version_info.major == 3


try:
    # Python 2
    from itertools import izip_longest
except ImportError:
    # Python 3
    from itertools import zip_longest as izip_longest


COLORPALLETTE = [
        '#00ff00', '#0000ff', '#00ffff', '#ff0000', '#ffff00', '#ff00ff',
        '#008000', '#000080', '#008080', '#00ff80', '#0080ff', '#800000', '#808000', '#80ff00',
        '#800080', '#808080', '#80ff80', '#8000ff', '#8080ff', '#80ffff', '#ff8000', '#ff0080',
        '#ff8080', '#ffff80', '#ff80ff',
        '#004000', '#00bf00', '#000040', '#004040', '#008040', '#00bf40', '#00ff40', '#004080',
        '#00bf80', '#0000bf', '#0040bf', '#0080bf', '#00bfbf', '#00ffbf', '#0040ff', '#00bfff',
        '#400000', '#404000', '#408000', '#40bf00', '#40ff00', '#400040', '#404040', '#408040',
        '#40bf40', '#40ff40', '#400080', '#404080', '#408080', '#40bf80', '#40ff80', '#4000bf',
        '#4040bf', '#4080bf', '#40bfbf', '#40ffbf', '#4000ff', '#4040ff', '#4080ff', '#40bfff',
        '#40ffff', '#804000', '#80bf00', '#800040', '#804040', '#808040', '#80bf40', '#80ff40',
        '#804080', '#80bf80', '#8000bf', '#8040bf', '#8080bf', '#80bfbf', '#80ffbf', '#8040ff',
        '#80bfff', '#bf0000', '#bf4000', '#bf8000', '#bfbf00', '#bfff00', '#bf0040', '#bf4040',
        '#bf8040', '#bfbf40', '#bfff40', '#bf0080', '#bf4080', '#bf8080', '#bfbf80', '#bfff80',
        '#bf00bf', '#bf40bf', '#bf80bf', '#bfbfbf', '#bfffbf', '#bf00ff', '#bf40ff', '#bf80ff',
        '#bfbfff', '#bfffff', '#ff4000', '#ffbf00', '#ff0040', '#ff4040', '#ff8040', '#ffbf40',
        '#ffff40', '#ff4080', '#ffbf80', '#ff00bf', '#ff40bf', '#ff80bf', '#ffbfbf', '#ffffbf',
        '#ff40ff', '#ffbfff'
    ]


def brightness(hexcode):
    """Calculates brightness for give html hex code of the format #xxxxxx"""
    return int(hexcode[1:3], 16) * .299 + int(hexcode[3:5], 16) * .587 + int(hexcode[5:7], 16) * .114

# Calculate brightness for each color and determine if text should be black or white.
FORMAT_COLORS = [(bg_color, '#000000' if brightness(bg_color) >= 128 else '#ffffff') for bg_color in COLORPALLETTE]


def grouper(n, iterable, fillvalue=None):
    """
    Groups iterable into n length chunks.
    If the last chunk doesn't have n items, the remaining is filled with fillvalue.

    >>> list(grouper(3, 'ABCDEFG', fillvalue='x'))
    [('A', 'B', 'C'), ('D', 'E', 'F'), ('G', 'x', 'x')]
    """
    args = [iter(iterable)] * n
    return izip_longest(fillvalue=fillvalue, *args)


def _iter_colors(data, color_map, default=None):
    """Yields byte and format color for each byte of data according to the member_map.

    :param data: Data to iterate over.
    :param color_map: Dictionary that matches offset a member and color to use
    :param default: default colors to use.

    :yield: tuple containing (byte, format_color_tuple)
    """
    iter_data = enumerate(data)
    for offset, datum in iter_data:
        if offset in color_map:
            colors, member = color_map[offset]
            yield datum, colors
            for offset, datum in itertools.islice(iter_data, 0, member.length - 1):
                yield datum, colors
        else:
            yield datum, default


class Member(construct.RawCopy):
    """
    This is a subconstruct that collects offset, data, and size information into the given
    member table, but then returns the original parsed value, like nothing happened.
    (This is to allow the callbacks work like they originally functioned.)
    """

    def __init__(self, member_map, subcon):
        """
        :param member_map: a defaultdict(list) mapping the offsets to the parsed objects
        :param subcon:
        """
        self._member_map = member_map
        super(Member, self).__init__(subcon)
        # version 2.9 doesn't perpetuate the name past one level anymore.
        self.name = self.subcon.name

    def _generate_value_str(self, value, indent=0):
        tabs = '\t' * indent
        if isinstance(value, construct.ListContainer):
            return '- ' + tabs + ('\n' + '- ' + tabs).join(
                self._generate_value_str(value_, indent=indent+1).lstrip()
                for value_ in value)
        elif isinstance(value, construct.Container):
            # NOTE: must use items() instead of iteritems() to keep order.
            return tabs + ('\n' + tabs).join(
                '{}: \n{}'.format(name, self._generate_value_str(value_, indent=indent+1))
                for name, value_ in value.items() if not name.startswith('_'))
        elif isinstance(value, bytes):
            # Escape unprintable bytes with "\x" notation.
            # (using codecs necessary to get this to work in both python 2 and 3)
            return tabs + codecs.escape_encode(value)[0].decode('utf-8')
        else:
            return tabs + '{}'.format(value)

    def _parse(self, stream, context, path):
        obj = super(Member, self)._parse(stream, context, path)

        # Store offset, data, and size information then return original object like nothing happened...
        if self.name and not self.name.startswith('_'):
            obj.name = self.name
            # Create a string representation of the value.
            obj.value_str = self._generate_value_str(obj.value)
            # Need path to so we can pull name history. (Add index if there is one.)
            obj.path = path
            index = getattr(context, '_index', None)
            if index is not None:
                obj.path += '[{}]'.format(index)
            obj.docs = self.docs

            # Map ourselves to every byte we cover.
            for index in range(obj.offset1, obj.offset2):
                self._member_map.setdefault(index, [])
                self._member_map[index].append(obj)

        return obj.value

    def _build(self, obj, stream, context, path):
        raise NotImplementedError('Unable to build using Member class.')


class MemberMap(construct.Adapter):
    r"""
    Wraps Subconstruct to produce a member map of all the parsed objections and their offsets:

    {offset: [list of parsed Containers in order of descending depth]}

    Needs to implement ``_decode()`` and ``_encode()``.

    :param subcon: the construct to wrap
    """
    def __init__(self, subcon):
        # member_map is a dictionary mapping the offsets of elements to a list of elements it portrays
        self._member_map = {}
        subcon = self._wrap_subcon(subcon)
        super(MemberMap, self).__init__(subcon)

    def _wrap_subcon(self, subcon):
        """Recursively wraps all subconstructs with Member."""
        # Don't wrap the Probes.
        if isinstance(subcon, construct.Probe):
            return subcon

        # Recursively wrap internals as until we hit an adapter or non-Construct object.
        if isinstance(subcon, construct.Construct) and not isinstance(subcon, construct.Adapter):
            if hasattr(subcon, 'subcon'):
                subcon.subcon = self._wrap_subcon(subcon.subcon)
            elif hasattr(subcon, 'subcons'):
                new_subcons = []
                for _subcon in subcon.subcons:
                    new_subcons.append(self._wrap_subcon(_subcon))
                subcon.subcons = new_subcons

        # Switch uses "cases"
        if isinstance(subcon, construct.Switch):
            new_cases = {}
            for case, _subcon in subcon.cases.items():
                new_cases[case] = self._wrap_subcon(_subcon)
            subcon.cases = new_cases
            subcon.default = self._wrap_subcon(subcon.default)

        return Member(self._member_map, subcon)

    def _parse(self, stream, context, path):
        # Clear the member_table from previous use.
        self._member_map.clear()
        return super(MemberMap, self)._parse(stream, context, path)

    def _decode(self, obj, context, path):
        """Returns a copy of the member map."""
        return self._member_map.copy()

    def _encode(self, obj, context, path):
        raise NotImplementedError('Not supported.')


def _gen_color_map(member_map, depth=1, member_callback=None):
    """
    Generates a color map that maps beginning offsets to a member.

    :param member_map: A dictionary map, mapping byte offsets to members.
    :param depth: The number of levels deep to display in table (defaults to all levels)
    :param member_callback: Optional callback that can be used to tweak the member name or value before setting.
    :return:
    """
    if depth is not None and depth <= 0:
        raise ValueError('Invalid depth. Must be >= 1 or None.')

    # Contains set of parent members that are not allowed to be present.
    # (This helps to prevent a parent being displayed when a child contains a unnamed member (e.g. Padding))
    blacklist = set()

    visible_members = {}
    curr_member = None
    for offset, members in sorted(member_map.items()):
        members = list(reversed(members))  # Members are generated in reverse with most depth being first.

        # Grab member based on requested level, (use furthest depth member if not requested)
        if depth is None:
            idx = len(members) - 1
        else:
            idx = min(depth - 1, len(members) - 1)
        member = members[idx]
        blacklist.update(id(m) for m in members[:idx])

        # Add to member to color map only if its the first time we are seeing it.
        if member != curr_member and id(member) not in blacklist:
            # Rename member to contain parent names.
            # ([1:] to remove the "parsing" name)
            member.name = ' / '.join(member.path.split(' -> ')[1:] + [member.name])
            if member_callback:
                results = member_callback(member.name, member.value_str)
                if results:
                    member.name, member.value_str = results
            visible_members[offset] = member
            curr_member = member

    # Remove any blacklisted members that slipped by.
    # (This happens when the first entry in a construct is empty.)
    visible_members = {offset: member
                       for offset, member in visible_members.items() if id(member) not in blacklist}

    # Assign colors.
    color_generator = itertools.cycle(FORMAT_COLORS)
    color_map = {offset: (next(color_generator), member) for offset, member in sorted(visible_members.items())}

    return color_map


def html_hex(struct, data, width=16, depth=None, member_callback=None):
    """
    Uses construct to parse data and creates a user-friendly html hex dump.

    :param struct: A construct.Construct object to parse.
    :param data: Data to dump.
    :param width: The number of bytes displayed for each line.
    :param depth: The number of levels deep to display in table (defaults to all levels)
    :param member_callback:
        Optional callback function that can be used to tweak the
        member name or value in the variable table.
        Function must accept two parameters (name, value) and return a tuple of
        the (name, value) or None to make no change.
        e.g.
        def edit_member(name, value):
            if name == 'data':
                return name, value.encode('hex')


    :rtype str: returns unicode string of html data.

    :raises ConstructError: If given struct fails to parse given data.
    """
    if width <= 0:
        raise ValueError("Width must be a positive number.")
    member_map = MemberMap(struct).parse(data)
    color_map = _gen_color_map(member_map, depth=depth, member_callback=member_callback)

    hex_dump = []
    for line_number, line in enumerate(grouper(width, _iter_colors(data, color_map), fillvalue=(None, None))):
        offset = line_number * width
        hex_line = []
        ascii_line = []
        # Generate hex and ascii version of each byte.
        current_color = None
        for byte, color in line:
            prefix, suffix = '', ''
            if color:
                if color != current_color:
                    bg_color, text_color = color
                    # End previous highlighting.
                    if current_color:
                        prefix += '</span>'
                    prefix += '<span style="background:{};color:{}">'.format(bg_color, text_color)
                    current_color = color
            # Clear highlighting.
            elif current_color:
                prefix = '</span>'
                current_color = None

            if byte is not None:
                if not PY3:
                    byte = ord(byte)
                hex_ = '{:02X}'.format(byte)
                ascii = chr(byte) if 32 < byte < 127 else '.'
            else:
                hex_ = '&nbsp;&nbsp;'
                ascii = '&nbsp;'

            hex_line.append('{}{}{}'.format(prefix, hex_, suffix))
            ascii_line.append('{}{}{}'.format(prefix, ascii, suffix))

        hex_line = '&nbsp;'.join(hex_line)
        ascii_line = ''.join(ascii_line)

        # Clear highlighting.
        if current_color:
            hex_line += '</span>'
            ascii_line += '</span>'

        hex_dump.append(('{:06x}'.format(offset), hex_line, ascii_line))

    env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(os.path.join(os.path.dirname(__file__))),
        trim_blocks=True, lstrip_blocks=True)
    template = env.get_template('construct_template.html')

    return template.render(hex_dump=hex_dump, color_map=color_map, width=width)


if __name__ == '__main__':
    # Run an example if called directly.
    from mwcp.utils.construct import version28 as construct
    from mwcp.utils.construct.network import IP4Address
    from mwcp.utils.construct.helpers import HexString
    from construct import this

    EMBED_SPEC = construct.Struct(
        'a' / IP4Address,
        'b' / IP4Address,
        'c' / IP4Address,
        'd' / IP4Address
    )

    address_struct = construct.Struct(
        'first' / construct.Struct('a' / construct.Byte, 'b' / construct.Byte),
        'second' / construct.Struct('inner2' / construct.Bytes(2))
        # 'internal' / IP4Address
    )

    PACKET = construct.Struct(
        construct.Padding(0x9),
        'Hardcoded Value 1' / HexString(construct.Int32ul),
        'Hardcoded Value 2' / HexString(construct.Int32ul),
        'Hardcoded Value 3' / HexString(construct.Int32ul),
        construct.Padding(0x17),
        'Compromised Host IP' / IP4Address,  # Use IP adapter
        # 'Unknown IP Addresses' / construct.Switch(
        #     this['Hardcoded Value 1'],
        #     {
        #         '0x1f4' : EMBED_SPEC
        #     },
        # ),
        'Unknown IP Addresses' / address_struct[4],
        # 'Unknown IP Addresses' / IP4Address[4],
        construct.Padding(8),
        'Unknown Indicator' / construct.String(0xF),
        construct.Padding(2),
        'Number of CPUs' / construct.Int32ul,
        'CPU Mhz' / construct.Int32ul,
        'Total Memory (MB)' / construct.Int32ul,
        'Compromised System Kernel' / construct.CString(),
        'Possible Trojan Version' / construct.CString()
    )

    data = (b'\x01\x00\x00\x00}\x00\x00\x00\x00\xf4\x01\x00\x002\x00\x00\x00\xe8'
            b'\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01'
            b'\x01\x00\x00\x00\x00\x01\x00\x00\x00\xc0\xa8\x01\r\xc0\xa8\x01\r\xc0'
            b'\xa8\x01\r\xc0\xa8\x01\r\xc0\xa8\x01\r\xff\xff\x01\x00\x00\x00\x00\x00'
            b'-== Love AV ==-:\x00\x01\x00\x00\x00d\n\x00\x00\xc4\x07\x00\x00'
            b'Linux 3.13.0-93-generic\x001:G2.40\x00')

    print(html_hex(PACKET, data, depth=2))




"""
Description: Utility for elftools python library.
"""

import logging

logger = logging.getLogger(__name__)

import elftools.elf.elffile as elffile
import io


def obtain_elf(file_data):
    """
    Given file data, create an elftools.ELFFile object from the data.

    :param file_data: Input ELF file data

    :return: An elftools.ELFFile object or None
    """
    try:
        elf = elffile.ELFFile(io.BytesIO(file_data))
        return elf
    except elffile.ELFError:
        logger.debug('An elftools.ELFFile object on the file data could not be created.')
        return None


def obtain_section(section_name, elf=None, file_data=None):
    """
    Obtain the section obtain for a specficied ELF section of a file.

    :param section_name: The name of the section to obtain
    :param elf: elftools.ELFFile object
    :param file_data: Input file data

    :return: The elftools.Section object, or None.
    """
    if file_data:
        elf = obtain_elf(file_data)
    if elf:
        for section in elf.iter_sections():
            if section.name == section_name:
                    return section
        return None
    else:
        return None


def obtain_section_data(section_name, elf=None, file_data=None, min_size=0):
    """
    Obtain the data in a specified ELF section of a file.

    :param section_name: The name of the section from which to extract data.
    :param elf: elftools.ELFFile object
    :param file_data: Input file data
    :param min_size: The minimum acceptable size for the section_data

    :return: The PE section data, or None.
    """
    if file_data:
        elf = obtain_elf(file_data)
    if elf:
        section = obtain_section(section_name, elf)
        if section:
            section_data = section.data()
            if len(section_data) > min_size:
                return section_data
            return None
        return None
    else:
        return None


def check_section(section_name, elf=None, file_data=None):
    """
    Check if a specified ELF section exists in a file.

    :param section_name: The name of the section from which to extract data.
    :param elf: elftools.ELFFile object
    :param file_data: Input file data

    :return: True if the section name is observed, False if it is not.
    """
    if file_data:
        elf = obtain_elf(file_data)
    if elf and obtain_section(section_name, elf):
        return True
    return False


def obtain_physical_offset(mem_offset, elf=None, file_data=None):
    """
    For an ELF file (in x86), convert a provided memory offset to a raw offset.

    :param mem_offset: The memory offset to convert to a raw offset
    :param elf: elftools.ELFFile object
    :param file_data: Input file data

    :return: Raw offset, or None.
    """
    if file_data:
        elf = obtain_elf(file_data)
    if elf:
        for phy_offset in elf.address_offsets(mem_offset):
            return phy_offset
    return None


def obtain_memory_offset(phy_offset, elf=None, file_data=None):
    """
    For an ELF file, convert a provided raw offset to a memory offset.

    :param phy_offset: The raw offset to convert to a memory offset
    :param elf: elftools.ELFFile object
    :param file_data: Input file data

    :return: Memory offset, or None.
    """
    if file_data:
        elf = obtain_elf(file_data)
    if elf:
        for seg in elf.iter_segments():
            if seg['p_offset'] <= phy_offset < (seg['p_offset'] + seg['p_filesz']):
                return phy_offset - seg['p_offset'] + seg['p_vaddr']
        return None
    else:
        return None
